/*****************************************************************************/

/*
 *      quasar.c -- NetStream2000 EM8400 (quasar) driver
 *
 *      Copyright (C) 1999-2000  Sigma Designs
 *                    written by Emmanuel Michon <emmanuel_michon@sdesigns.com>
 *                           and Howard G. Page
 *
 *      2.3.x/2.4.x compatibility by Thomas Sparr <thomas.sparr@kreatel.se>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*****************************************************************************/

#include <linux/version.h>

#include <linux/pci.h>
#include <linux/wrapper.h>
#include <linux/pagemap.h>

#include <asm/io.h>
#include <asm/uaccess.h>

#include "common.h"
#include "quasar.h"
#include "realmagicdev.h"

// look in linux/fs.h for types
// int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
// int (*mmap) (struct file *, struct vm_area_struct *);
// int (*open) (struct inode *, struct file *);
// int (*release) (struct inode *, struct file *);

int quasar_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int quasar_mmap(struct file *, struct vm_area_struct *);
int quasar_open(struct inode *, struct file *);
int quasar_release(struct inode *, struct file *);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
// so fun to change the order in function arrays.
struct file_operations quasar_fops = 
{
	NULL, // struct module *owner;
	NULL, // loff_t (*llseek) (struct file *, loff_t, int);
	NULL, // ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	NULL, // ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
	NULL, // int (*readdir) (struct file *, void *, filldir_t);
	NULL, // unsigned int (*poll) (struct file *, struct poll_table_struct *);
	quasar_ioctl, // int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
	quasar_mmap, // int (*mmap) (struct file *, struct vm_area_struct *);
	quasar_open, // int (*open) (struct inode *, struct file *);
	NULL, // int (*flush) (struct file *);
	quasar_release, // int (*release) (struct inode *, struct file *);
	NULL, // int (*fsync) (struct file *, struct dentry *, int datasync);
	NULL, // int (*fasync) (int, struct file *, int);
	NULL, // int (*lock) (struct file *, int, struct file_lock *);
	NULL, // ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *);
	NULL // ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *);
};
#else
// struct file_operations has 15 members for linux 2.2.12, 14, 16, 17.
static struct file_operations quasar_fops =
{
    NULL,              /* llseek function.     */
    NULL,              /* read function.       */
    NULL,              /* write function.      */
    NULL,              /* read directory.      */
    NULL,              /* poll function.       */
    quasar_ioctl,      /* ioctl function.      */
    quasar_mmap,       /* mmap function.       */
    quasar_open,       /* open function.       */
    NULL,              /* flush function.      */
    quasar_release,    /* release function.    */
    NULL,              /* sync file func.      */
    NULL,              /* async file func.     */
    NULL,              /* media change check.  */
    NULL,              /* revalidate function. */
    NULL               /* lock function.       */
};
#endif

// the same routine handles all REALmagic devices
void quasar_interrupt(int irq, void *dev_t, struct pt_regs * regs);

/* convert virtual user memory address to physical address */
/* (unlike virt_to_phys for kernel virtual space memory) */
/* taken from btty.c */
static inline unsigned long uvirt_to_phys(unsigned long adr)
{
        unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;
        pgd_t *pgd = pgd_offset(current->mm, adr);
  
	if (!pgd_none(*pgd)) {
                pmd = pmd_offset(pgd, adr);
                if (!pmd_none(*pmd)) {
                        ptep = pte_offset(pmd, adr);
                        pte = *ptep;
                        if (pte_present(pte)) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,23)
                                unsigned long kva;
                                kva = ((unsigned long)(page_address(pte_page(pte)))
				       |((unsigned long)(adr&(PAGE_SIZE-1))));
                                ret = virt_to_phys((void*) kva);
#else
                                ret = virt_to_phys((void*) (pte_page(pte)|(adr&(PAGE_SIZE-1))));
#endif
                        }
                }
        }
	return ret;
}


/* Pure 2^n version of get_order */
static inline int my_get_order(size_t size)
{
        unsigned long order;

        size = (size-1) >> (PAGE_SHIFT-1);
        order = -1;
        do {
                size >>= 1;
                order++;
        } while (size);
        return order;
}

static inline void *dmaalloc(size_t size)
{
	unsigned long addr = __get_dma_pages(GFP_KERNEL, my_get_order(size));

	if (!addr) printk("QUASAR: __get_dma_pages %d %d failed!\n",size, my_get_order(size));
	else 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		unsigned long page;

		for (page=addr;page<addr+size;page+=PAGE_SIZE)
			SetPageReserved(virt_to_page(page));
	}
#else
	{
		int i;
		
		for (i = MAP_NR(addr); i < MAP_NR(addr+size); i++) 
			mem_map_reserve(i);
	}
#endif
    
	return (void *)addr;
}

static inline void dmafree(void *addr, size_t size) 
{
	printk("dmafree called with %p, %d\n",addr,size);

	if (size>0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		unsigned long page;

		for (page=(unsigned long)addr;page<(unsigned long)addr+size;page+=PAGE_SIZE)
			ClearPageReserved(virt_to_page(page));
	}
#else
	{
		int i;
		
		for (i = MAP_NR((unsigned long)addr);
		     i < MAP_NR((unsigned long)addr+size); i++) {
			mem_map_unreserve (i);
		}
	}
#endif		
		free_pages((unsigned long) addr, my_get_order(size));
	}
}

int quasar_ioctl(struct inode *i_node, struct file *filp,
		 unsigned int cmd, unsigned long arg)
{
	int idx = MINOR(i_node->i_rdev)&REALMAGIC_MINOR_DEV_MASK;
	realmagic_dev_t *dev = &realmagic_dev_tab[idx];
	
	switch (cmd) {
		
	case IOCTL_GET_PIOMAPSIZE:
		{
			LOGINFO(1,"ioctl device #%d IOCTL_GET_PIOMAPSIZE successful",idx);
			copy_to_user((void *)arg,&dev->mapsize,sizeof(int));
			return 0;
		}

	case IOCTL_DMABUF_ALLOC:
		{
			int i;

			for (i=0;i<MAX_dmabuf;i++) if (!dev->dmabuf_table[i].inuse) break;
			if (i==MAX_dmabuf) {
				printk("MAX_dmabuf reached\n");
				return -EINVAL;
			}

			DUMPDMABUF(IOCTL_DMABUF_ALLOC)
			
			copy_from_user(&dev->dmabuf_table[i],(void *)arg,sizeof(dmabuf_helper));

			DUMPDMABUF(IOCTL_DMABUF_ALLOC)

			dev->dmabuf_table[i].KernelVirtualAddress=
				(unsigned long)dmaalloc(dev->dmabuf_table[i].Size);
			dev->dmabuf_table[i].PhysicalAddress=
				virt_to_phys((void *)dev->dmabuf_table[i].KernelVirtualAddress);
			dev->dmabuf_table[i].inuse=1;
			dev->dmabuf_table[i].ProcessVirtualAddress=0;

			DUMPDMABUF(IOCTL_DMABUF_ALLOC)

			copy_to_user((void *)arg,&dev->dmabuf_table[i],sizeof(dmabuf_helper));
			return 0;
		}
			
	case IOCTL_DMABUF_FREE:
		{
			/* the user space code must call munmap and put zero in ProcessVirtualAddress 
			   field before getting here */

			int i;
			dmabuf_helper d;

			copy_from_user(&d,(void *)arg,sizeof(dmabuf_helper));
			
			// looking for the (only) index using KernelVirtualAddress
			for (i=0;i<MAX_dmabuf;i++) 
				if (dev->dmabuf_table[i].inuse&&
				    (dev->dmabuf_table[i].KernelVirtualAddress==d.KernelVirtualAddress)) break;
			if (i==MAX_dmabuf) {
				printk("MAX_dmabuf reached\n");
				return -EINVAL;
			}

			DUMPDMABUF(IOCTL_DMABUF_FREE)
				
			if (d.ProcessVirtualAddress) {
				printk("Trying to dmafree mmapped memory\n");
				return -EINVAL;
			}
			else {
				dmafree((void *)dev->dmabuf_table[i].KernelVirtualAddress,dev->dmabuf_table[i].Size);
				dev->dmabuf_table[i].inuse=0;
			}
			
			DUMPDMABUF(IOCTL_DMABUF_FREE)
			
			copy_to_user((void *)arg,&dev->dmabuf_table[i],sizeof(dmabuf_helper));
		}
			
	case IOCTL_TRANSLATE_PROCESSVIRTUAL_TO_PHYSICAL:
		{
			// the arg is the # of the dmabuf
			unsigned long processvirtual,physical;
			copy_from_user(&processvirtual,(void *)arg,sizeof(unsigned long));
			
			physical = uvirt_to_phys(processvirtual);
			// printk("translating processvirtual %lx to physical %lx\n",processvirtual,physical);
			
			copy_to_user((void *)arg,&physical,sizeof(unsigned long));
			return 0;
		}
		
	case IOCTL_SEND_PID:
		{
			copy_from_user(&dev->mpegsrv_pid,(void *)arg,sizeof(unsigned long));
			printk("[Received user space IH pid %ld]\n",dev->mpegsrv_pid);
			break;
		}

	case IOCTL_SEND_IRQ_STATUS_REGISTER:
		{
			copy_from_user(&dev->irq_status_register,(void *)arg,sizeof(unsigned long));
			printk("[Received irq status register address at %08lx]\n",dev->irq_status_register);
			break;
		}
		
	default:
		return -EINVAL;
	}
	
	return 0; // never reached anyways
}

int quasar_mmap(struct file * filp, struct vm_area_struct * vma)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,25)
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
#else
        unsigned long offset = vma->vm_offset;
#endif
	
	int idx = MINOR(filp->f_dentry->d_inode->i_rdev)&REALMAGIC_MINOR_DEV_MASK;
	realmagic_dev_t *dev = &realmagic_dev_tab[idx];

	switch (offset) {
	case REMAP_PIO:
		{
			// for PIO space
			VERBOSECALL(DEBUG,remap_page_range(vma->vm_start,dev->ioaddr, 
							   vma->vm_end - vma->vm_start, 
							   vma->vm_page_prot),
				    -EINVAL,RAS);
			
			LOGINFO(1,"mmap (remap_page_range) device #%d successful",idx);
			
			return 0;
		}

	case REMAP_DMA:
		{
			// for DMA space

			int i;

			// looking for the (only) dmabuf_table entry used but not yet mapped 
			for (i=0;i<MAX_dmabuf;i++) 
				if (dev->dmabuf_table[i].inuse&!dev->dmabuf_table[i].ProcessVirtualAddress) break;
			if (i==MAX_dmabuf) {
				printk("MAX_dmabuf reached\n");
				return -EINVAL;
			}
			
			DUMPDMABUF(REMAP_DMA)
				
			VERBOSECALL(DEBUG,remap_page_range(vma->vm_start,
							   dev->dmabuf_table[i].PhysicalAddress,
							   vma->vm_end - vma->vm_start, 
							   vma->vm_page_prot),
				    -EINVAL,RAS);	

			dev->dmabuf_table[i].ProcessVirtualAddress=vma->vm_start;

			DUMPDMABUF(REMAP_DMA)

			return 0;
		}
		
	default:
		return -EINVAL;
	}
	
	return 0; // never reached anyways
}

int quasar_open(struct inode *i_node, struct file *filp)
{
	int idx = MINOR(i_node->i_rdev)&REALMAGIC_MINOR_DEV_MASK;
	realmagic_dev_t *dev = &realmagic_dev_tab[idx];
	int i;

	LOGINFO(1,"opening device #%d successful (doing nothing)",
		MINOR(i_node->i_rdev)&REALMAGIC_MINOR_DEV_MASK);
	
	// not a good idea to open twice.
	if (dev->opentimes!=0) {
		printk("quasar_open: already open\n");
		return -EINVAL;
	}
	
	dev->opentimes++;
	dev->mpegsrv_pid=-1;
	dev->irq_status_register=-1;
	
	LOGINFO(1,"opening device #%d successful (doing nothing)",
		MINOR(i_node->i_rdev)&REALMAGIC_MINOR_DEV_MASK);
	
	// clean dmabuf_table
	for (i=0;i<MAX_dmabuf;i++) dev->dmabuf_table[i].inuse=0;

	return 0;
}

int quasar_release(struct inode *i_node, struct file *filp)
{
	int idx = MINOR(i_node->i_rdev)&REALMAGIC_MINOR_DEV_MASK;
	realmagic_dev_t *dev = &realmagic_dev_tab[idx];

	LOGINFO(1,"releasing device #%d successful (doing nothing)",
		MINOR(i_node->i_rdev)&REALMAGIC_MINOR_DEV_MASK);

	// not a good idea to close twice.
	if (dev->opentimes==0) {
		printk("quasar_release: already closed\n");
		return -EINVAL;
	}
	
	dev->opentimes--;

	if (dev->opentimes==0) {	
		int i;

		for (i=0;i<MAX_dmabuf;i++) 
			if (dev->dmabuf_table[i].inuse) {
				
				DUMPDMABUF(release)

				if (dev->dmabuf_table[i].ProcessVirtualAddress) 
					printk("Trying to dmafree mmapped memory --- doing it anyways\n");
				dmafree((void *)dev->dmabuf_table[i].KernelVirtualAddress,dev->dmabuf_table[i].Size);
				dev->dmabuf_table[i].inuse=0;

				DUMPDMABUF(release)

			}
	}

	return 0;
}

int pci_scan(void)
{
	struct pci_dev *pdev = NULL;
	int pcipresent;
	
	// check for PCI BIOS
	pcipresent=pci_present();
	ASSERT(DEBUG,pcipresent,
	       -EINVAL,RAS);
	
	// look for Sigma devices
	realmagic_dev_idx=0;

	while ((pdev=pci_find_device(PCI_VENDOR_ID_SIGMAD,
				     PCI_DEVICE_ID_SIGMAD_PCICARD,pdev))) {
		realmagic_dev_t *dev;
		
		ASSERT(DEBUG,(realmagic_dev_idx<REALMAGIC_MAX_DEV),
		       -EINVAL,RAS);

		dev=&realmagic_dev_tab[realmagic_dev_idx];
		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,13)
		dev->ioaddr = pdev->resource[0].start;
#else
                dev->ioaddr = pdev->base_address[0];
#endif	

		dev->irq    = pdev->irq;		
		dev->irqcount=0;

		dev->fops   = &quasar_fops;

		// to be cleaned up
		{
			int tmp;
			unsigned short v;
			
			pci_read_config_dword(pdev,PCI_BASE_ADDRESS_0,&tmp);
			pci_write_config_dword(pdev,PCI_BASE_ADDRESS_0,~0);
			pci_read_config_dword(pdev,PCI_BASE_ADDRESS_0,&dev->mapsize);
			dev->mapsize=~dev->mapsize+1;
			pci_write_config_dword(pdev,  PCI_BASE_ADDRESS_0, tmp);

			pcibios_read_config_word(pdev->bus->number,
						 pdev->devfn,PCI_COMMAND,&v);
			LOGINFO(1,"cmd=0x%hx",v);
			pcibios_read_config_word(pdev->bus->number, 
						 pdev->devfn, PCI_STATUS,&v);
			LOGINFO(1,"stat=0x%hx",v);
		}
		
		LOGINFO(1,"found REALmagic device #%d at PCI "
			"slot=%d devfn=%d ioaddr=%08lx mapsize=%d irq=%d",
			realmagic_dev_idx,PCI_SLOT(pdev->devfn), 
			pdev->devfn,dev->ioaddr,dev->mapsize,dev->irq);
		
		realmagic_dev_idx++;
	}
	
	ASSERT(1,realmagic_dev_idx,
	       -EINVAL,"found %d REALmagic devices total",realmagic_dev_idx);
	
	return 0;
}

int quasar_init(void)
{
	int idx;

	VERBOSECALL(DEBUG,pci_scan(),
		    -EINVAL,RAS);
	
	for(idx=0;idx<realmagic_dev_idx;idx++) {
		realmagic_dev_t *dev=&realmagic_dev_tab[idx];

		// ioremap
		dev->vioaddr=(unsigned long)ioremap(dev->ioaddr&PAGE_MASK,dev->mapsize);
		ASSERT(1,dev->vioaddr,
		       -EINVAL,"ioremap device #%d successful at %08lx",idx,dev->vioaddr);
		
		// irq request. ``dev'' is unique: allows irq sharing
		VERBOSECALL(DEBUG,request_irq(dev->irq,quasar_interrupt,
					      SA_SHIRQ,INTERRUPT_ENTRY_NAME,dev),
			    -EINVAL,"ok for irq %d on device #%d",dev->irq,idx);

		// relevant values in struct...
		// init of dmabuf_table is done in open method
		dev->opentimes=0;
	}
	
	return 0;
}

int quasar_cleanup(void)
{
	int idx,nok=0;
	
	// free interrupt
	for(idx=0;idx<realmagic_dev_idx;idx++) {
		realmagic_dev_t *dev=&realmagic_dev_tab[idx];

		if (dev->opentimes==0) {	
			// iounmap
			iounmap((void*)(dev->vioaddr&PAGE_MASK));
			
			LOGINFO(1,"iounmap device #%d executed (no return code)",idx);
			
			// irq free
			free_irq(dev->irq,dev);
			
			// FIXME CODE TO FREE LAST BUFFERS

			nok++;
		}
		else printk("quasar_cleanup: instance #%d still open\n",idx);
	}
	
	if (nok<realmagic_dev_idx) return -EINVAL;
	
	return 0;
}

void quasar_interrupt(int irq, void *dev_t, struct pt_regs * regs)
{
#define CLEAR_IRQ_REGISTER 0x1FFA
#define CLEAR_IRQ 0x2

	static int otherirqcount=0;
        realmagic_dev_t* dev = (realmagic_dev_t*)dev_t;
	unsigned long *clearirq=(unsigned long *)(dev->vioaddr+(CLEAR_IRQ_REGISTER<<2));

	/* in case of irq sharing, quick way out.
	   if we do not have irq_status_register yet, that's not a problem. */
	if (dev->irq_status_register!=-1) {
		unsigned long *irqstatus=(unsigned long *)(dev->vioaddr+(dev->irq_status_register<<2));

#define IRQ_STATUS 0x8000
		
		if ((*irqstatus&IRQ_STATUS)==0) {
			otherirqcount++;
			if ((otherirqcount%10000)==0) printk("[Skipped %d interrupts from other device sharing line]\n",otherirqcount);
			
			return;
		}
	}

	dev->irqcount++;
        *clearirq = CLEAR_IRQ;
        
        if (dev->mpegsrv_pid!=-1) { 
		// with 5000 should log about every minute
		if (dev->irqcount%5000==0) printk("[Sent %d interrupts so far to pid %ld]\n",dev->irqcount,dev->mpegsrv_pid);
		
		// I do not know clearly what is the last arg. Maybe the
		// origin of the signal but now, it's the kernel, so, what 
		// should we put there?
                kill_proc(dev->mpegsrv_pid,SIGUSR2,0);
        }
	
        // do not reenable interrupt: user mode does it for us.
        return;
}
