/* NSC_LIC_BSD
 *
 * National Semiconductor Corporation Open Source License for
 *
 *    Video library for Geode SCx2x0
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) 2000-2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use by Licensee of this Software in source and
 * binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must either:
 *     (a) reproduce the above copyright notice, this list of conditions
 *         and the following disclaimer in the documentation and/or
 *         other materials provided with the distribution, or
 *     (b) if Licensee elects to redistribute any Software not in
 *         accordance with (a), then Licensee shall indemnify National
 *         Semiconductor Corporation and all copyright holders and
 *         contributors for any claims related to such Software (and pay
 *         all costs and damages, including all attorney fees)
 *
 *   * Neither the name of the National Semiconductor Corporation nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF
 * YOUR JURISDICTION. It is Licensee's responsibility to comply with
 * any export regulations applicable in Licensee's jurisdiction. Under
 * CURRENT (July 2001) U.S. export regulations this software
 * is eligible for export from the U.S. and can be downloaded by or
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran,
 * Syria, Sudan, Afghanistan and any other country to which the U.S.
 * has embargoed goods and services.
 *
 * END_NSC_LIC_BSD */
 
/*--------------------------------------------------------------------------
 * Copyright (c) 2000-2001 by National Semiconductor Corporation
 * All rights reserved.
 *--------------------------------------------------------------------------
 * File Contents:
 *           OS dependent functions needed by the SCx2x0 Video Library,
 *           implemented for the VxWorks and Linux OSs.
 * Project : Video library for Geode SCx2x0
 * Created : November, 2000
 * Author  : Galit Heller & Eyal Cohen
 *--------------------------------------------------------------------------
 * $Date: 8/10/01 6:58a $
 * $RCSfile: vl_os.c $
 * $ProjectRevision: 3.1 $
 * $Revision: 6 $
 *------------------------------------------------------------------------*/

/*========================================================================*/
/*========================================================================*/
/*============================   VxWorks OS   ============================*/
/*========================================================================*/
/*========================================================================*/

#ifdef OS_VXWORKS

/*------------------------------------------------------------------------*/
/*-------------------------   Include files   ----------------------------*/
/*------------------------------------------------------------------------*/
#include <vxWorks.h>
#include <intLib.h>
#include <iv.h>
#include <drv/pci/pciConfigLib.h>
#include <sysLib.h>
#include <rngLib.h>
#include "vl_rtns.h"
#include "vl_os.h"

/*------------------------------------------------------------------------*/
/*----------------   Constants and macros definitions  -------------------*/
/*------------------------------------------------------------------------*/
#define INT_NUM_IRQ0 0x20

/*------------------------------------------------------------------------*/
/*---------------------   Data Types definitions   -----------------------*/
/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*----------------------   Variable definitions   ------------------------*/
/*------------------------------------------------------------------------*/

/*-------------*/
/*-- Externs --*/
/*-------------*/


/**************************************************************************/
/************************   Interrupt services   **************************/
/**************************************************************************/


//---------------------------------------------------------------------
// Description:
//   Connect an interrupt handler to an IRQ and enable it.
//
// Parameters:
//   IRQL     - The IRQ level.
//   function - The interrupt handler to be connected.
//   int_name - The name of the interrupt.
//   param    - The parameter to be passed to the interrupt handler.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS OS_InterruptConnect (int IRQL, void (* function)(int), const char *int_name, int param)
{
	if (intConnect(INUM_TO_IVEC(INT_NUM_IRQ0 + IRQL), (VOIDFUNCPTR)function, param) != OK)
		return VL_STATUS_ERROR;
	if (sysIntEnablePIC (IRQL) == OK)
		return VL_STATUS_OK;
	else
		return VL_STATUS_ERROR;
}

//---------------------------------------------------------------------
// Description:
//   Disconnect the interrupt handler from the required IRQ.
//
// Parameters:
//   IRQL - The IRQ level.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS OS_InterruptDisconnect (int IRQL)
{
	return VL_STATUS_OK;
}


/**************************************************************************/
/***********************   RingBuffer services   **************************/
/**************************************************************************/


//---------------------------------------------------------------------
// Description:
//   Creates an empty ring buffer of size "nbytes", and initializes it.
//
// Parameters:
//   nbytes - The size of the ring buffer in bytes.
//
// Return Value:
//   The ID of the ring buffer, or NULL if memory cannot be allocated.
//---------------------------------------------------------------------
VL_RING_ID OS_RingCreate (int nbytes)
{
	return (VL_RING_ID) rngCreate(nbytes);
}

//---------------------------------------------------------------------
// Description:
//   Delete the specified ring buffer.
//
// Parameters:
//   ringId - The ID of the ring buffer.
//
// Return Value:
//   None.
//
// Side Effects:
//   Any data currently in the buffer will be lost.
//---------------------------------------------------------------------
void OS_RingDelete (VL_RING_ID ringId)
{
	rngDelete((RING_ID)ringId);
}

//---------------------------------------------------------------------
// Description:
//   Get bytes from the specified ring buffer into the specified buffer.
//   It copies as many bytes as are available in the ring buffer,
//   up to the specified maximum number.
//   The bytes copied will be removed from the ring buffer.
//
// Parameters:
//   ringId   - The ID of the ring buffer.
//   buffer   - The buffer to be filled with data from the ring buffer.
//   maxbytes - The maximum number of bytes to copy.
//
// Return Value:
//   The number of bytes actually copied from the ring buffer.
//   It may be zero if the ring buffer is empty at the time of the call. 
//---------------------------------------------------------------------
int OS_RingBufGet (VL_RING_ID ringId, char *buffer, int maxbytes)
{
	return rngBufGet((RING_ID)ringId, buffer, maxbytes);
}

//---------------------------------------------------------------------
// Description:
//   Put bytes from a specified buffer into the specified ring buffer.
//   The specified number of bytes will be put into the ring buffer,
//   up to the number of bytes available in the ring buffer.
//
// Parameters:
//   ringId   - The ID of the ring buffer.
//   buffer   - The buffer from where to take the data.
//   maxbytes - The number of bytes to copy.
//
// Return Value:
//   The number of bytes actually put into the ring buffer.
//   It may be less than the number requested, even zero, if there is
//   insufficient room in the ring buffer at the time of the call. 
//---------------------------------------------------------------------
int OS_RingBufPut (VL_RING_ID ringId, char *buffer, int nbytes)
{
	return rngBufPut((RING_ID)ringId, buffer, nbytes);
}


/**************************************************************************/
/**************************   PCI services   ******************************/
/**************************************************************************/

//---------------------------------------------------------------------
// Description:
//   Read/Write PCI Byte/Word/DWord.
//
// Parameters:
//   bus     - Bus number
//   dev     - Device number
//   func    - Function number
//   address - Offset in the configuration space
//   data    - Data read/written
//
// Return Value:
//
//---------------------------------------------------------------------

VL_STATUS OS_PciReadByte (int bus, int dev, int func, int address, unsigned char * data)
{
	if (pciConfigInByte(bus, dev, func, address, data) == OK)
		return VL_STATUS_OK;
	else
		return VL_STATUS_ERROR;
}

VL_STATUS OS_PciWriteByte (int bus, int dev, int func, int address, unsigned char data)
{
	if (pciConfigOutByte(bus, dev, func, address, data) == OK)
		return VL_STATUS_OK;
	else
		return VL_STATUS_ERROR;
}

VL_STATUS OS_PciReadWord (int bus, int dev, int func, int address, unsigned short * data)
{
	if (pciConfigInWord(bus, dev, func, address, data) == OK)
		return VL_STATUS_OK;
	else
		return VL_STATUS_ERROR;
}

VL_STATUS OS_PciWriteWord (int bus, int dev, int func, int address, unsigned short data)
{
	if (pciConfigOutWord(bus, dev, func, address, data) == OK)
		return VL_STATUS_OK;
	else
		return VL_STATUS_ERROR;
}

VL_STATUS OS_PciReadDWord (int bus, int dev, int func, int address, unsigned long * data)
{
	if (pciConfigInLong(bus, dev, func, address, (unsigned int *)data) == OK)
		return VL_STATUS_OK;
	else
		return VL_STATUS_ERROR;
}

VL_STATUS OS_PciWriteDWord (int bus, int dev, int func, int address, unsigned long data)
{
	if (pciConfigOutLong(bus, dev, func, address, (unsigned int)data) == OK)
		return VL_STATUS_OK;
	else
		return VL_STATUS_ERROR;
}


/**************************************************************************/
/*************************   I/O port access   ****************************/
/**************************************************************************/

//---------------------------------------------------------------------
// Description:
//   Input/Output Byte/Word/DWord from/to I/O space.
//
// Parameters:
//   port - I/O port address
//   data - Data read/written from/to the port
//
// Return Value:
//
//---------------------------------------------------------------------

unsigned char OS_IoReadByte (unsigned short port)
{
	return sysInByte(port);
}

void OS_IoWriteByte (unsigned short port, unsigned char data)
{
	sysOutByte(port, (char) data);
}

unsigned short OS_IoReadWord (unsigned short port)
{
	return sysInWord(port);
}

void OS_IoWriteWord (unsigned short port, unsigned short data)
{
	sysOutWord(port, (short) data);
}

unsigned long OS_IoReadDWord (unsigned short port)
{
	return sysInLong(port);
}

void OS_IoWriteDWord (unsigned short port, unsigned long data)
{
	sysOutLong(port, (long) data);
}

#endif /* OS_VXWORKS */


/*========================================================================*/
/*========================================================================*/
/*=============================   Linux OS   =============================*/
/*========================================================================*/
/*========================================================================*/

#ifdef OS_LINUX

#include <asm/io.h> 
#include <linux/malloc.h> /* kmalloc */
#include <linux/pci.h>
#include "vl_rtns.h"
#include "vl_os.h"

typedef struct VLhandler_t {
    void (* func)(int);
    int param;
} VLhandler_t;

typedef struct RingBuf_t {
    unsigned long put_ptr;
    unsigned long get_ptr;
    int           buf_size;
    char *        buffer;
} RingBuf_t;


static VLhandler_t VL_Handler[16];

/**************************************************************************/
/************************   start and stop OS services   ******************/
/**************************************************************************/
//---------------------------------------------------------------------
// Description:
//   Initialize this OS library. Call this function before using
//   any of the OS services.
//
// Parameters:
//   None.
//
// Return Value:
//   None.
//---------------------------------------------------------------------
void OS_Interface_init (void)
{
    memset(VL_Handler, 0, sizeof(VLhandler_t) * 16);
}

//---------------------------------------------------------------------
// Description:
//   Free all resources which were used by this OS library. Call this
//   function before unloading (or removing) the driver.
//
// Parameters:
//   None.
//
// Return Value:
//   None.
//---------------------------------------------------------------------
void OS_Interface_close (void)
{
    int irq;
    /* Free all used IRQs */
    for(irq = 0 ; irq < 16 ; irq++) {
        if (VL_Handler[irq].func) {
            free_irq(irq, &(VL_Handler[irq]));
        }
    }
}

/**************************************************************************/
/************************   Interrupt services   **************************/
/**************************************************************************/

static void linux_handler (int irq, void *dev_id, struct pt_regs *regs)
{
    VLhandler_t * VL_Handler = (VLhandler_t *)dev_id;
    VL_Handler->func(VL_Handler->param);
}


//---------------------------------------------------------------------
// Description:
//   Connect an interrupt handler to an IRQ.
//
// Parameters:
//   IRQL     - The IRQ level.
//   function - The interrupt handler to be connected.
//   param    - The parameter to be passed to the interrupt handler.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS OS_InterruptConnect (int IRQL, void (* function)(int), const char *int_name, int param)
{
    VL_Handler[IRQL].func = function;
    VL_Handler[IRQL].param = param;
    if (request_irq(IRQL, &linux_handler, 0, int_name, &(VL_Handler[IRQL])))
 		return VL_STATUS_ERROR;
	else
		return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Disconnect the interrupt handler from the required IRQ.
//
// Parameters:
//   IRQL - The IRQ level.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS OS_InterruptDisconnect (int IRQL)
{
	if (VL_Handler[IRQL].func)
		free_irq(IRQL, &(VL_Handler[IRQL]));
	VL_Handler[IRQL].func = 0;
	return VL_STATUS_OK;
}


/**************************************************************************/
/***********************   RingBuffer services   **************************/
/**************************************************************************/

/*
   Following is a simple implementation of a ring buffer.
   It might not be the best implementation and it also has a bug when
   put_ptr or get_ptr reaches MAXINT (0xFFFFFFFF).
   But, for the Video Library demo, it is good enough since this library uses
   the ring buffer for CC data, which comes in a very slow rate, at about
   100 bytes per second. So with this rate it will take put_ptr more than
   one year to reach 0xFFFFFFFF.
*/

//---------------------------------------------------------------------
// Description:
//   Creates an empty ring buffer of size "nbytes", and initializes it.
//
// Parameters:
//   nbytes - The size of the ring buffer in bytes.
//
// Return Value:
//   The ID of the ring buffer, or NULL if memory cannot be allocated.
//---------------------------------------------------------------------
VL_RING_ID OS_RingCreate (int nbytes)
{
	RingBuf_t * ring = (RingBuf_t *)kmalloc(sizeof(RingBuf_t), GFP_KERNEL);
	ring->put_ptr  = 0;
	ring->get_ptr  = 0;
	ring->buf_size = nbytes;
	ring->buffer   = (char *) kmalloc(nbytes, GFP_KERNEL);
	memset(ring->buffer, 0, nbytes);
	return (VL_RING_ID) ring;
}

//---------------------------------------------------------------------
// Description:
//   Delete the specified ring buffer.
//
// Parameters:
//   ringId - The ID of the ring buffer.
//
// Return Value:
//   None.
//
// Side Effects:
//   Any data currently in the buffer will be lost.
//---------------------------------------------------------------------
void OS_RingDelete (VL_RING_ID ringId)
{
	RingBuf_t * ring = ringId;
	kfree(ring->buffer);
	kfree(ring);
}

//---------------------------------------------------------------------
// Description:
//   Get bytes from the specified ring buffer into the specified buffer.
//   It copies as many bytes as are available in the ring buffer,
//   up to the specified maximum number.
//   The bytes copied will be removed from the ring buffer.
//
// Parameters:
//   ringId   - The ID of the ring buffer.
//   buffer   - The buffer to be filled with data from the ring buffer.
//   maxbytes - The maximum number of bytes to copy.
//
// Return Value:
//   The number of bytes actually copied from the ring buffer.
//   It may be zero if the ring buffer is empty at the time of the call. 
//---------------------------------------------------------------------
int OS_RingBufGet (VL_RING_ID ringId, char *buffer, int nbytes)
{
	RingBuf_t * ring = ringId;
	int i;
	for(i = 0 ; i < nbytes ; i ++ ) {
		if (ring->get_ptr == ring->put_ptr)
			return i;
		buffer[i] = ring->buffer [ring->get_ptr % ring->buf_size];
		ring->get_ptr ++;
	}
	return i;
}

//---------------------------------------------------------------------
// Description:
//   Put bytes from a specified buffer into the specified ring buffer.
//   The specified number of bytes will be put into the ring buffer,
//   up to the number of bytes available in the ring buffer.
//
// Parameters:
//   ringId   - The ID of the ring buffer.
//   buffer   - The buffer from where to take the data.
//   maxbytes - The number of bytes to copy.
//
// Return Value:
//   The number of bytes actually put into the ring buffer.
//   It may be less than the number requested, even zero, if there is
//   insufficient room in the ring buffer at the time of the call. 
//---------------------------------------------------------------------
int OS_RingBufPut (VL_RING_ID ringId, char *buffer, int nbytes)
{
	RingBuf_t * ring = ringId;
	int i;
	for(i = 0 ; i < nbytes ; i ++ ) {
		if (ring->put_ptr - ring->get_ptr >= ring->buf_size)
			return i;
		ring->buffer [ring->put_ptr % ring->buf_size] = buffer[i];
		ring->put_ptr ++;
	}
	return i;
}


/**************************************************************************/
/**************************   PCI services   ******************************/
/**************************************************************************/

//---------------------------------------------------------------------
// Description:
//   Read/Write PCI Byte/Word/DWord.
//
// Parameters:
//   bus     - Bus number
//   dev     - Device number
//   func    - Function number
//   address - Offset in the configuration space
//   data    - Data read/written
//
// Return Value:
//
//---------------------------------------------------------------------

VL_STATUS OS_PciReadByte (int bus, int dev, int func, int address, unsigned char * data)
{
	if (pcibios_read_config_byte ((unsigned char)bus, (unsigned char) PCI_DEVFN(dev, func),
                                  (unsigned char)address, data))
		return VL_STATUS_ERROR;
	else
		return VL_STATUS_OK;
}

VL_STATUS OS_PciWriteByte (int bus, int dev, int func, int address, unsigned char data)
{
	if (pcibios_write_config_byte ((unsigned char)bus, (unsigned char) PCI_DEVFN(dev, func),
                                   (unsigned char)address, data))
		return VL_STATUS_ERROR;
	else
		return VL_STATUS_OK;
}

VL_STATUS OS_PciReadWord (int bus, int dev, int func, int address, unsigned short * data)
{
	if (pcibios_read_config_word ((unsigned char)bus, (unsigned char) PCI_DEVFN(dev, func),
                                  (unsigned char)address, data))
		return VL_STATUS_ERROR;
	else
		return VL_STATUS_OK;
}

VL_STATUS OS_PciWriteWord (int bus, int dev, int func, int address, unsigned short data)
{
	if (pcibios_write_config_word ((unsigned char)bus, (unsigned char) PCI_DEVFN(dev, func),
                                   (unsigned char)address, data))
		return VL_STATUS_ERROR;
	else
		return VL_STATUS_OK;
}

VL_STATUS OS_PciReadDWord (int bus, int dev, int func, int address, unsigned long * data)
{
	if (pcibios_read_config_dword ((unsigned char)bus, (unsigned char) PCI_DEVFN(dev, func),
                                   (unsigned char)address, (unsigned int *)data))
		return VL_STATUS_ERROR;
	else
		return VL_STATUS_OK;
}

VL_STATUS OS_PciWriteDWord (int bus, int dev, int func, int address, unsigned long data)
{
	if (pcibios_write_config_dword ((unsigned char)bus, (unsigned char) PCI_DEVFN(dev, func),
                                    (unsigned char)address, (unsigned int)data))
		return VL_STATUS_ERROR;
	else
		return VL_STATUS_OK;
}


/**************************************************************************/
/*************************   I/O port access   ****************************/
/**************************************************************************/

//---------------------------------------------------------------------
// Description:
//   Input/Output Byte/Word/DWord from/to I/O space.
//
// Parameters:
//   port - I/O port address
//   data - Data read/written from/to the port
//
// Return Value:
//
//---------------------------------------------------------------------

unsigned char OS_IoReadByte (unsigned short port)
{
	return inb(port);
}

void OS_IoWriteByte (unsigned short port, unsigned char data)
{
	outb(data, port);
}

unsigned short OS_IoReadWord (unsigned short port)
{
	return inw(port);
}

void OS_IoWriteWord (unsigned short port, unsigned short data)
{
	outw(data, port);
}

unsigned long OS_IoReadDWord (unsigned short port)
{
	return inl(port);
}

void OS_IoWriteDWord (unsigned short port, unsigned long data)
{
	outl(data, port);
}

#endif /* OS_LINUX */
