/* NSC_LIC_BSD
 *
 * National Semiconductor Corporation Open Source License for
 *
 *    Video library for Geode SCx2x0
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) 2000-2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use by Licensee of this Software in source and
 * binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must either:
 *     (a) reproduce the above copyright notice, this list of conditions
 *         and the following disclaimer in the documentation and/or
 *         other materials provided with the distribution, or
 *     (b) if Licensee elects to redistribute any Software not in
 *         accordance with (a), then Licensee shall indemnify National
 *         Semiconductor Corporation and all copyright holders and
 *         contributors for any claims related to such Software (and pay
 *         all costs and damages, including all attorney fees)
 *
 *   * Neither the name of the National Semiconductor Corporation nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF
 * YOUR JURISDICTION. It is Licensee's responsibility to comply with
 * any export regulations applicable in Licensee's jurisdiction. Under
 * CURRENT (July 2001) U.S. export regulations this software
 * is eligible for export from the U.S. and can be downloaded by or
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran,
 * Syria, Sudan, Afghanistan and any other country to which the U.S.
 * has embargoed goods and services.
 *
 * END_NSC_LIC_BSD */
 
/*--------------------------------------------------------------------------
 * Copyright (c) 2000-2001 by National Semiconductor Corporation
 * All rights reserved.
 *--------------------------------------------------------------------------
 * File Contents:
 *           Functions for video initialization, control, configuration for
 *           display on CRT/TV, and termination. 
 * Project : Video library for Geode SCx2x0
 * Created : November, 2000
 * Author  : Galit Heller & Eyal Cohen
 *--------------------------------------------------------------------------
 * $Date: 9/25/01 12:21a $
 * $RCSfile: vl_vid.c $
 * $ProjectRevision: 3.1 $
 * $Revision: 9 $
 *------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*-------------------------   Include files   ----------------------------*/
/*------------------------------------------------------------------------*/
#include "gfx_rtns.h"
#include "gfx_defs.h"
#include "gfx_regs.h"
#include "gfx_mode.h"
#include "vl_rtns.h"
#include "vl_defs.h"
#include "vl_os.h"

/*------------------------------------------------------------------------*/
/*----------------   Constants and macros definitions  -------------------*/
/*------------------------------------------------------------------------*/
#undef TIMER_DEBUG

// GPIO definitions
#define GPIO_SELECT					0x6420
#define GPIO_CONFIG					0x6424
#define GPIO_BANK_OFFSET			5
#define GPIO_PME_RISING_EDGE		0x20
#define GPIO_PME_FALLING_EDGE		0x00

// Timer definitions
#define ONE_SEC						32768
#define TRIGGER_DIFF				(-4)
#define SETTLE_INTERVAL				(VidTimings.RefreshRate/4)
#define VERIFY_INTERVAL				(VidTimings.RefreshRate/10)

// Gfx timing definitions
#define GFX_TIMING_NEG_HSYNC         0x1
#define GFX_TIMING_NEG_VSYNC         0x2
#define GFX_TIMING_LOCK              0x1000

/*------------------------------------------------------------------------*/
/*--------------------   Data Types definitions   ------------------------*/
/*------------------------------------------------------------------------*/

typedef enum VL_GfxTVResolution {
    VL_GFX_TV_RESOLUTION_NTSC,
    VL_GFX_TV_RESOLUTION_PAL,
    VL_GFX_TV_RESOLUTION_NTSC_NO_SCALING,
    VL_GFX_TV_RESOLUTION_PAL_NO_SCALING,
} VL_GfxTVResolution;

/*------------------------------------------------------------------------*/
/*----------------------   Variable definitions   ------------------------*/
/*------------------------------------------------------------------------*/

// In the following array each structure represents the timings for a
// specific screen format. The entries must be arranged in the same
// order as the formats in "VL_GfxTVResolution" type.

DISPLAYMODE GfxDisplayTimings[] = {
/* NTSC resolution */
{ GFX_TIMING_NEG_HSYNC | GFX_TIMING_NEG_VSYNC | GFX_TIMING_LOCK,
  640, 640, 656, 744, 792, 792,             /* horizontal timings   */
  480, 480, 490, 492, 517, 525,	            /* vertical timings     */
  0x0018EC4D,                               /* freq = 24.923052 MHz */
},
/* PAL resolution */
{ GFX_TIMING_NEG_HSYNC | GFX_TIMING_NEG_VSYNC | GFX_TIMING_LOCK,
  768, 768, 800, 848, 864, 864,	            /* horizontal timings   */
  576, 576, 586, 588, 625, 625,	            /* vertical timings     */
  0x001B0000,                               /* freq = 27.00 MHz     */
},
/* NTSC resolution non-square pixels */
{ GFX_TIMING_NEG_HSYNC | GFX_TIMING_NEG_VSYNC | GFX_TIMING_LOCK,
  720, 720, 736, 752, 792, 792,	            /* horizontal timings   */
  480, 480, 490, 492, 517, 525,	            /* vertical timings     */
  0x0018EC4D,                               /* freq = 24.923052 MHz */
},
/* PAL resolution non-square pixels */
{ GFX_TIMING_NEG_HSYNC | GFX_TIMING_NEG_VSYNC | GFX_TIMING_LOCK,
  720, 720, 752, 816, 864, 864,             /* horizontal timings   */
  576, 576, 586, 588, 625, 625,             /* vertical timings     */
  0x001B0000,                               /* freq = 27.00 MHz     */
}};

VidTimingsChar  VidTimings;
VidDisplayState VidDisplay;
VL_VideoStruct  VidAttr;

int timer_count;
int field_count;
int fix = 1;
int vip_field;
int line_diff;
int sig_count;
unsigned short gpioBank;
unsigned short gpioNum;
unsigned short time_till_checking_status; // let tv encoder status stabilize
unsigned long inverse_field_count, inverse_field_fix;
unsigned long same_field_count, same_field_fix;

#ifdef DEBUG

typedef struct sync {
	unsigned short vip_line;
	unsigned short gxm_line;
	unsigned short vip_field;
	unsigned short tvenc_field;
} sync;

#define SYNC_BUF_SIZE 40
sync sync_buf[SYNC_BUF_SIZE];

unsigned long sync_buf_counter = 0;
unsigned long tearing_counter = 0;
unsigned long field_sync_counter = 0;

#endif // DEBUG

const char *vl_version = "$ProjectRevision: 3.1 $";
char VidLibVer[10];
static boolean VideoInitialized = FALSE;

/*-------------*/
/*-- Externs --*/
/*-------------*/
extern boolean using_memory_temporarily;
extern boolean change_drift_mode;
extern unsigned long vip_even_base, vip_odd_base;
extern VL_MemBuf VidDisplayBuf;
extern VL_MemBuf VidCaptureBuf;

/**************************************************************************/
/************************   Interrupt handlers   **************************/
/**************************************************************************/


/*
Timer interrupt :
    Triggered at the middle of the video overlay (e.g., graphics
    line 240 if full screen NTSC is used).
    The interrupt handler is responsible for the following things :
    - Extract CC information from the ring buffer and program it in
      the TV encoder registers.
    - Toggle video buffers for interlaced VIP, i.e., tell GXm which
      field to send to TVOUT in the next field. This is done by programming
      the video offset in the GXm (register 0x8320).
    - Execute periodic algorithms like video tearing correction for scaled
      video windows.

Signal monitoring interrupt :
    Triggered when the video decoder detects a bad signal.
    In DVIP mode genlock to the input signal is active.
	When the input signal becomes not stable from some reason,
	genlock to the input signal must be disabled in order to be able
	to continue displaying a stable graphics image.
    If the video decoder in the system is able to provide signal status
	information, this can be used to disable genlock to the input signal
	and switch from DVIP to VIP.
*/

//========================================================
//===================    TIMER    ========================
//========================================================

int time_to_middle_of_overlay()
{
	int numlines;
	int vtotal = gfx_get_vtotal();
	int middle = VidDisplay.current_window_y + VidDisplay.current_window_height/2;
	if (middle < 40) middle = 40;
	else if (middle > 440) middle = 440;
	numlines = vtotal - gfx_get_vline() + middle;
	return (ONE_SEC * numlines / VidTimings.RefreshRate / vtotal);
}

static void timerIntHandler()
{
	unsigned short tvenc_field, vip_line, gxm_line;
	int prev_vip_field;
	int prev_line_diff;
	int signal_ok;
	boolean toggle_tv_field = 0;
	
	// Clear watchdog overflow
	OS_IoWriteByte(SC1200_CB_BASE_ADDR + SC1200_CB_WDSTS, 1);

	// Config WDTO to end at the middle of overlay on the next frame
	OS_IoWriteWord(SC1200_CB_BASE_ADDR + SC1200_CB_WDTO, time_to_middle_of_overlay());
	timer_count++;

#ifdef TIMER_DEBUG
	OS_IoWriteByte(0x80, timer_count);
#endif

	// Save previous field data
	prev_vip_field = vip_field;
	prev_line_diff = line_diff;
	
	// Get status
	vip_field = gfx_test_vip_odd_field();
	vip_line = (unsigned short)READ_VIP32(SC1200_VIP_CURRENT_LINE);
	gxm_line = gfx_get_vline();
	tvenc_field = gfx_test_tvenc_odd_field();
	
	line_diff = gxm_line / 2 - vip_line;

#ifdef DEBUG
	sync_buf[sync_buf_counter].vip_line = vip_line;
	sync_buf[sync_buf_counter].gxm_line = gxm_line;
	sync_buf[sync_buf_counter].vip_field = vip_field;
	sync_buf[sync_buf_counter].tvenc_field = tvenc_field;
	if (++sync_buf_counter == SYNC_BUF_SIZE) sync_buf_counter = 0;
#endif

	// Video signal status monitoring (based on field types)
	// to choose between VIP and DVIP

	if (VidDisplay.current_video_source == VL_VIDEO_SOURCE_DVIP) {
		if (vip_field == prev_vip_field) {
			field_count++;
			if (field_count == 3)
				VidSetDirectVIP(0, 0xC1);
		}
		else field_count = 0;
	} else if ((VidDisplay.current_window_size == VL_VIDEO_WINDOW_FULL) &&
	           (VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP) && !VidDisplay.paused) {
		if ((vip_field != prev_vip_field) &&
			(line_diff - prev_line_diff < 5) &&
			(line_diff - prev_line_diff > -5)) {
			field_count++;
			if (field_count >= 10) {
				if ((line_diff == TRIGGER_DIFF) && (vip_field == tvenc_field)) {
					if (VidAttr.SignalMonitorEnable) {
						if (gpioBank == 0)
							signal_ok = OS_IoReadDWord(0x6404) & 0x80;
						else
							signal_ok = OS_IoReadDWord(0x6414) & 0x80;
					}
					else signal_ok = 1;
					if (signal_ok && fix)
						VidSetDirectVIP(1, 0xD1);
					else field_count = 0;
				}
			}
		}
		else field_count = 0;
	}
	
	// Program CC data in the TV encoder
	VidSendCCDataToTVenc();

	if (time_till_checking_status) time_till_checking_status--;

	// SYNC FIELD TYPE OF INPUT (VIP) TO OUTPUT (TV) WHEN DVIP IS ACTIVE

	if (((VidDisplay.current_video_source == VL_VIDEO_SOURCE_DVIP) ||
		 (VidDisplay.current_video_source == VL_VIDEO_SOURCE_DVIP_ONLY)) &&
		!time_till_checking_status &&
		(VidAttr.TVStandard != VL_TV_STANDARD_NONE))
	{
		if (tvenc_field != vip_field) {
			inverse_field_count++;
			if (inverse_field_count == inverse_field_fix) {
				toggle_tv_field = TRUE;
#ifdef DEBUG
				field_sync_counter++;
#endif
			}
		}
		else inverse_field_count = 0;
	}
	else inverse_field_count = 0;

	// TEARING CORRECTION
	if ((VidDisplay.tearing_correction_active) && (vip_line != 0) &&
		!time_till_checking_status && !using_memory_temporarily)
	{
		if (tvenc_field == vip_field) {
			same_field_count++;
			if (same_field_count == same_field_fix) {
				toggle_tv_field = TRUE;
#ifdef DEBUG
				tearing_counter++;
#endif
			}
		}
		else same_field_count = 0;
	}
	else same_field_count = 0;

	// TOGGLE TV FIELD IF NEEDED
	if (toggle_tv_field)
	{
		int vphase = gfx_get_tv_vphase();
		if (vphase > 20)
			gfx_set_tv_vphase(vphase - VPHASE_OFFSET);
		else
			gfx_set_tv_vphase(vphase + VPHASE_OFFSET);
		if ((VidDisplay.current_window_size == VL_VIDEO_WINDOW_SCALED) &&
			gfx_get_genlock_enable())
			// Reset TV encoder at the next odd field to prevent drift
			// This is needed/allowed only when in genlock mode
			gfx_set_tvenc_reset_interval(TVENC_RESET_NEXT_ODD_FIELD);
		// Allow toggle effect to settle before checking the need for another one
		time_till_checking_status = SETTLE_INTERVAL;
		inverse_field_count = 0;
		same_field_count = 0;
	}

	// HANDLE VIDEO BUFFER TOGGLING
	if ((VidDisplay.current_video_source == VL_VIDEO_SOURCE_MEMORY) &&
		(VidDisplay.current_deinterlacing_method != VL_VIDEO_DEINTERLACING_INTERLEAVE) &&
		!VidDisplay.paused)
		{
		if (tvenc_field == 0)
		{
			// Even field being shown now, odd field will be shown on next frame
			gfx_set_video_offset(VidDisplay.memory_address_odd - (unsigned long)gfx_phys_fbptr);
		} else {
			// Odd field being shown now, even field will be shown on next frame
			gfx_set_video_offset(VidDisplay.memory_address_even - (unsigned long)gfx_phys_fbptr);
		}
	}					
}

void timer_stop(void)
{
	OS_IoWriteWord(SC1200_CB_BASE_ADDR + SC1200_CB_WDTO, 0);

	// Disconnect the timer interrupt handler
	OS_InterruptDisconnect(VidAttr.TimerIRQ);
}

VL_STATUS timer_start(void)
{
	timer_count = 0;
	field_count = 0;
	time_till_checking_status = 0;
	inverse_field_count = 0;
	inverse_field_fix = VERIFY_INTERVAL;
	same_field_count = 0;
	same_field_fix = VERIFY_INTERVAL;

#ifdef DEBUG
	sync_buf_counter = 0;
	tearing_counter = 0;
	field_sync_counter = 0;
#endif DEBUG	

	// Configure High-Resolution Timer: Oscillator clock divided by 27; Interrupts disabled
	OS_IoWriteByte(SC1200_CB_BASE_ADDR + SC1200_CB_TMCNFG, 0);

	// Config INTSEL - select IRQ for the watchdog
	OS_IoWriteByte(SC1200_CB_BASE_ADDR + SC1200_CB_INTSEL, VidAttr.TimerIRQ);

	// Config WDTYPE1 to generate an interrupt
	OS_IoWriteWord(SC1200_CB_BASE_ADDR + SC1200_CB_WDCNFG, 0x10);

	// Clear watchdog overflow
	OS_IoWriteByte(SC1200_CB_BASE_ADDR + SC1200_CB_WDSTS, 1);

	// Connect the interrupt service routine
	if (OS_InterruptConnect(VidAttr.TimerIRQ, (void (*)(int))timerIntHandler, "VL_Timer", 0) != VL_STATUS_OK)
		return VL_STATUS_ERROR;

	// Config WDTO to end at the middle on the next frame
	OS_IoWriteWord(SC1200_CB_BASE_ADDR + SC1200_CB_WDTO, time_to_middle_of_overlay());

	return VL_STATUS_OK;
}

//========================================================
//=================== End of TIMER =======================
//========================================================

//========================================================
//===============  Signal Detection ======================
//========================================================

void signalISR()
{
	/* Clear status bits */
	if (gpioBank == 0)
		OS_IoWriteDWord(0x640C, (OS_IoReadDWord(0x640C) | 0x80));
	else
		OS_IoWriteDWord(0x641C, (OS_IoReadDWord(0x641C) | 0x80));
	OS_IoWriteWord(0xAC10, (OS_IoReadWord(0xAC10) | 0x08));

	// Check if appropriate situation to switch from DVIP to VIP
	if ((VidDisplay.full_screen_source != VL_VIDEO_SOURCE_DVIP) ||
		(VidDisplay.current_video_source == VL_VIDEO_SOURCE_MEMORY))
	return;
	
	sig_count++;

	// Switch to VIP
	VidSetDirectVIP(0, sig_count);
}

void SignalDetectionEnd(void)
{
	// Disconnect the signal monitoring interrupt handler
	OS_InterruptDisconnect(VidAttr.SignalMonitorIRQ);
}

// F0BAR0 (F0+0x10) = GPIO = 0x6400
// F1BAR1 (F1+0x40) = SCI  = 0xAC00
// Configure the requested GPIO as input pin to get SAA7114 signal status
VL_STATUS SignalDetectionInit(int gpio)
{
	unsigned long  tmpl;
	unsigned short tmps;

	if ((gpio >= 0) && (gpio <=15)) {
		gpioBank = 0;
		gpioNum = gpio;
	}
	else if ((gpio >= 32) && (gpio <=47)) {
		gpioBank = 1;
		gpioNum = gpio - 32;
	}
	else return VL_STATUS_ERROR;

	// For GPIO7 configure PMR[17] = 0, PMR[8] = 0
	if (gpio == 7) {
		tmpl = OS_IoReadDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR);
		tmpl &= ~0x20100;
		OS_IoWriteDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR, tmpl);
	}

	// Configure the requested GPIO as input, PME falling edge
	OS_IoWriteDWord(GPIO_SELECT, (gpioBank << GPIO_BANK_OFFSET) + gpioNum);
	OS_IoWriteDWord(GPIO_CONFIG, GPIO_PME_FALLING_EDGE);

	// Enable power management events for the requested GPIO
	if (gpioBank == 0)
		OS_IoWriteDWord(0x6408, 0x80);
	else
		OS_IoWriteDWord(0x6418, 0x80);

	// Allow GPIOs to generate an SCI
	tmps = OS_IoReadWord(0xAC12);
	tmps |= 0x8;
	OS_IoWriteWord(0xAC12, tmps);

	// Route the requested GPIO to IRQ10
	tmpl = OS_IoReadDWord(0xAC18);
	tmpl &= (~0xf);
	tmpl |= VidAttr.SignalMonitorIRQ;
	OS_IoWriteDWord(0xAC18, tmpl);

	// Enable PME to generate SCI
	tmps = OS_IoReadWord(0xAC0C);
	tmps |= 0x1;
	OS_IoWriteWord(0xAC0C, tmps);

	// Clear the requested GPIO status bit
	if (gpioBank == 0)
		OS_IoWriteDWord(0x640C, (OS_IoReadDWord(0x640C) | 0x80));
	else
		OS_IoWriteDWord(0x641C, (OS_IoReadDWord(0x641C) | 0x80));

	// Clear GPIO status (PME was caused by GPIO)
	OS_IoWriteWord(0xAC10, (OS_IoReadWord(0xAC10) | 0x08));

	// Connect the interrupt service routine for time_based field detection
	if (OS_InterruptConnect (VidAttr.SignalMonitorIRQ, (void (*)(int))signalISR, "VL_SignalMonitor", 0) != VL_STATUS_OK)
		return VL_STATUS_ERROR;

	sig_count = 0;

	return VL_STATUS_OK;
}

//========================================================
//============  End of Signal Detection ==================
//========================================================

/**************************************************************************/
/*******************   Initialization / Termination   *********************/
/**************************************************************************/


//---------------------------------------------------------------------
// Description:
//	 Initializes the video path by configuring the VIP,
//	 GXm and Video Processor modules according to the values specified.
//
//   This function must be called prior to using any of the other
//   video functions.
//   It is the responsibility of the caller to allocate the graphics frame
//   buffer according to the required TV Standard.
//
//   * Detect the type and revison of the CPU and video modules.
//   * Configure Display Controller.
//   * Configure TV output according to the specified standard and output type.
//   * Initialize VIP :
//	 - Set mode C
//	 - Disable video capture
//	 - Disable VBI capture
//   * Initialize video path :
//   - Set format to UYVY
//   - Scale video to compensate for the below mentioned aspect ratio scaling
//   - Initialize the Gamma palette with linear values and configure video
//     stream to pass through it
//   - Initialize color key to full magenta
//   - Select the graphics stream to be compared to the color key
//   - Disable graphics overlay on video, i.e., no graphics will be
//     displayed inside the video window
//   - Disable video cursor
//   - Display video outside the alpha windows, i.e., color key comparison
//     is not performed outside the Alpha window
//   - Disable all alpha windows
//   - Maintain correct graphics aspect ratio on TV, i.e.,
//     scale blended graphics and video image by the aspect ratio for
//     the requested TV standard
//   * Set the following preferences :
//   - Set video de-interlacing method for full and scaled video to Bob.
//   - Set full screen video to DVIP (if possible).
//   - Disable CC passing thru to the TV.
//   * Configure, connect and enable the following interrupts according to the parameters:
//   - Timer interrupt
//   - Signal monitoring
//
// Parameters:
//     video_struct fields:
//     TVStandard - The TV standard to be used, and to which the TV output,
//                  display controller will be configured.
//				    In case no TV output is required, set to NONE, and
//				    no change will be made to display controller timing
//				    or video color space.
//                  If "TVStandard" is NTSC :
//                  Frame buffer active size is : 640x480 pixels (draw only to the
//	                safe area since TV viewable area is less than the active).
//                  If "TVStandard" is PAL : same sequence as for NTSC but with PAL settings.
//                  Frame buffer active size is : 768x576 pixels (draw only to the
//	                safe area since TV viewable area is less than the active).
//     TVOutput   - The TV output type to be used. Does not need to be specified
//                  if TV is not enabled.
//     CRTEnable  - If TRUE, the function takes into consideration that a CRT
//                  is used in the system: DVIP cannot be used, video color
//                  space is set to RGB, and the CRT is enabled. 
//                  If FALSE, YUV color space is used, genlock is used (for both VIP
//                  and DVIP), and the CRT is disabled.
//     FullScreenSource
//                - The video source to be used for full screen (for scaled windows
//                  VIP is used).
//                  VL_VIDEO_SOURCE_DVIP_ONLY : only DVIP.
//                  VL_VIDEO_SOURCE_MEMORY :    only VIP.
//                  VL_VIDEO_SOURCE_DVIP : DVIP when signal is stable, otherwise VIP.
//     GenlockDelay
//                - The genlock delay value to be used when DVIP is active.
//     FieldPolarityInvert
//                 - TRUE if input video device inverts the field polarity status.
//     TimerIRQ   - The IRQ level to be used for the timer interrupt, which handles
//                  periodic video maintenance.
//     SignalMonitorEnable -
//                  TRUE :  the video signal monitoring interrupt handler should be
//                          activated. This interrupt handler utilizes the decoder's video
//                          signal detection capability.
//                          This requires a specific connection between the decoder and
//                          the Geode chip.
//                  FALSE : the video signal monitoring interrupt handler should not
//                          be activated.
//     SignalMonitorGPIONum
//                - The GPIO of the Geode chip to which the decoder's signal detection
//                  output signal is connected.
//                  Does not need to be specified if SignalMonitorEnable is FALSE.
//     SignalMonitorIRQ
//                - The IRQ level to be used for the signal monitoring interrupt.
//                  Does not need to be specified if SignalMonitorEnable is FALSE.
//     TVAspectRatio
//                Controls video and graphics aspect ratios on TV. Does not need to be
//                specified if TV is not enabled.
//                VL_TV_ASPECT_RATIO_GRAPHICS_AND_VIDEO
//                  The graphics will be scaled to appear with correct aspect
//                  ratio on TV.
//                  Video is prescaled to compensate for the aspect ratio
//                  scaling.
//                VL_TV_ASPECT_RATIO_VIDEO_ONLY
//                  No scaling is done to graphics, which means it will not
//                  appear in square pixels as on CRT
//                VL_TV_ASPECT_RATIO_GRAPHICS_ONLY
//                  The graphics will be scaled to appear with correct aspect
//                  ratio on TV.
//                  Video is not prescaled. The application should provide
//                  the video scaled by the reciprocal of the aspect ratio
//                  scaling.
//     Width, Height
//                - Filled by the called function to indicate to the caller the video
//                  source resolution.
//     Members of GfxStruct:
//     GfxAspectRatio
//                - The factor by which graphics is multiplied horizontally to
//                  achieve correct aspect ratio on TV.
//     GfxWidth   - Filled by the called function to indicate to the caller the
//                  width (in pixels) of the graphics frame buffer after configuring
//                  the video per the requested parameters.
//     GfxHeight  - Filled by the called function to indicate to the caller the
//                  height (in lines) of the graphics frame buffer after configuring
//                  the video per the requested parameters.
//     GfxPitch   - Filled by the called function to indicate to the caller the
//                  pitch (in bytes) of the graphics frame buffer after configuring
//                  the video per the requested parameters.
//   When VL_VideoInit() returns, the video is not enabled.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VideoInit(VL_VideoStruct * video_struct)
{
	unsigned char delayed_transactions;
	VL_STATUS ret_val;
	int hz;
	
	if (VideoInitialized) {
		 printk("VideoInit: Already initialized\n");
		 return VL_STATUS_ERROR;
	}
	VidAttr = *video_struct;

	// Disable genlock to prevent display controller timing stuck
	gfx_set_genlock_enable(0);
	
	// INITIALIZE VidDisplay

	// Determine FULL SCREEN SOURCE
	if (VidAttr.CRTEnable)
		VidDisplay.full_screen_source = VL_VIDEO_SOURCE_MEMORY;
	else
		VidDisplay.full_screen_source = VidAttr.FullScreenSource;

	// Initializations according to FULL SCREEN SOURCE
	if (VidAttr.FullScreenSource == VL_VIDEO_SOURCE_DVIP_ONLY) {
		VidDisplay.current_video_source = VL_VIDEO_SOURCE_DVIP_ONLY;
		if (VidAttr.TVStandard == VL_TV_STANDARD_NTSC)
			VidDisplay.extra_y_offset = EXTRA_Y_FOR_DVIP_NTSC;
		else
			VidDisplay.extra_y_offset = EXTRA_Y_FOR_DVIP_PAL;					
	} else {
		VidDisplay.current_video_source = VL_VIDEO_SOURCE_MEMORY;
		VidDisplay.extra_y_offset = 0;
	}
	if ((gfx_chip_revision >= SC1200_REV_C1) && (!VidAttr.CRTEnable)) {
		VidDisplay.full_screen_deinterlacing = VL_VIDEO_DEINTERLACING_NONE;
		VidDisplay.scaled_video_deinterlacing = VL_VIDEO_DEINTERLACING_NONE;
		VidDisplay.current_deinterlacing_method = VL_VIDEO_DEINTERLACING_BOB;  // Force first time setting
	} else {
		VidDisplay.full_screen_deinterlacing = VL_VIDEO_DEINTERLACING_BOB;
		VidDisplay.scaled_video_deinterlacing = VL_VIDEO_DEINTERLACING_BOB;
		VidDisplay.current_deinterlacing_method = VL_VIDEO_DEINTERLACING_NONE;  // Force first time setting
	}		
	VidDisplay.tearing_correction_active = FALSE;
	VidDisplay.current_window_size = VL_VIDEO_WINDOW_SCALED;             // Start from scaled to enter DVIP via VIP
	VidDisplay.horzFactor.num = VidDisplay.horzFactor.denom = 1;
	VidDisplay.paused = 0;

	VidDisplayBuf.address = 0;

	// INITIALIZE VidTimings

	switch (VidAttr.TVStandard) {
		case VL_TV_STANDARD_NTSC:
			VidTimings.Width = video_struct->Width = 720;
			VidTimings.Height = video_struct->Height = 480;
			VidTimings.VerticalTotal = 525;
			VidTimings.RefreshRate = 60;
			VidTimings.LineTime = 0x6b4; // Tim.htotal x 2 x 27 MHz / 24.923052 MHz
			break;
		case VL_TV_STANDARD_PAL:
			VidTimings.Width = video_struct->Width = 720;
			VidTimings.Height = video_struct->Height = 576;
			VidTimings.VerticalTotal = 625;
			VidTimings.RefreshRate = 50;
			VidTimings.LineTime = 0x6c0; // Tim.htotal x 2 x 27 MHz / 27 MHz
			break;
		case VL_TV_STANDARD_NONE:
			VidTimings.Width = video_struct->Width = 0;   // Unknown
			VidTimings.Height = video_struct->Height = 0; // Unknown
			video_struct->GfxStruct.GfxWidth = gfx_get_hactive();
			video_struct->GfxStruct.GfxHeight = gfx_get_vactive();
			video_struct->GfxStruct.GfxPitch = gfx_get_display_pitch();
			gfx_set_tv_format(TV_STANDARD_NTSC, GFX_ON_TV_SQUARE_PIXELS); // Needed for Bob on CRT
			VidTimings.VerticalTotal = 525;
			gfx_get_refreshrate_from_frequency(gfx_get_htotal(), gfx_get_vtotal(),
			(int)gfx_get_display_bpp(), &hz, gfx_get_clock_frequency());
			VidTimings.RefreshRate = (unsigned short)hz;
			VidTimings.LineTime = 0;       // TBD when implementing DVIP on CRT
			break;
		default:
			printk("VideoInit : Bad parameter : %d\n",__LINE__);
			return VL_STATUS_BAD_PARAMETER;
	}

	// INITIALIZE VIP
	gfx_set_vip_mode(VIP_MODE_C);
	gfx_set_video_source(VIDEO_SOURCE_MEMORY);
	gfx_set_vip_base(vip_even_base, vip_odd_base);
	gfx_set_vip_pitch(0);

	// VIP FIFO issues a bus request when it is filled with 32 bytes
	gfx_set_vip_bus_request_threshold_high(0);

	// Enable PCI Delayed Transactions for AT legacy PIC I/O Addresses
	OS_PciReadByte(0, 0x12, 0, 0x43, &delayed_transactions);
	delayed_transactions |= 0x4;
	OS_PciWriteByte(0, 0x12, 0, 0x43, delayed_transactions);

	// Enable capture to memory
	gfx_set_vip_enable(0);
	gfx_set_vip_capture_run_mode(VIP_CAPTURE_START_FIELD);

	// INITIALIZE VIDEO OVERLAY HARDWARE
	// Initialize scale to 1:1
	gfx_set_video_upscale(VidTimings.Width, VidTimings.Height,
		                  VidTimings.Width, VidTimings.Height);
	gfx_set_video_downscale_enable(0);		                  
	gfx_set_video_format(VIDEO_FORMAT_UYVY);
	if (VidAttr.CRTEnable)
		gfx_set_color_space_YUV(0); // actually done by gfx_set_video_format
	else
		gfx_set_color_space_YUV(1);
	gfx_set_video_filter(0, 0);
	gfx_set_video_palette((unsigned long *)0);
	gfx_set_video_cursor(0xffffff, 0xffffff, 0, 0, 0);  // Disable cursor match

	// Set Video Request values
	gfx_set_video_request(-46, -33);

	// Set color key to full magenta and color mask to 5:6:5
	VL_VideoSetColorKey(0x00F800F8, 0x00F8FCF8);

	// Disable graphics overlay on video
	VL_VideoSetOverlay(VL_VIDEO_OVERLAY_ALWAYS_ON);

	// Initialize video processing settings
	VidSetDeinterlacingMethod();
	VL_FieldPolarityInvert(VidAttr.FieldPolarityInvert);

	// TV CONFIGURATION

	if (VidAttr.TVStandard != VL_TV_STANDARD_NONE)
	{
		// SET TV OUTPUT

		switch (VidAttr.TVOutput) {
			case VL_TV_OUTPUT_COMPOSITE:
				gfx_set_tv_output(TV_OUTPUT_COMPOSITE);
				break;
			case VL_TV_OUTPUT_S_VIDEO:
				gfx_set_tv_output(TV_OUTPUT_S_VIDEO);
				break;
			case VL_TV_OUTPUT_YUV:
				gfx_set_tv_output(TV_OUTPUT_YUV);
				break;
			case VL_TV_OUTPUT_SCART:
				gfx_set_tv_output(TV_OUTPUT_SCART);
				break;
			default:
				printk("VideoInit: bad parameter : %d\n",__LINE__);
				return VL_STATUS_BAD_PARAMETER;
		}
		// CONFIGURE GRAPHICS AND VIDEO ASPECT RATIOS ON TV

		ret_val = VL_TVAspectRatio(VidAttr.TVAspectRatio, &video_struct->GfxStruct);
		if (ret_val != VL_STATUS_OK) {
			printk("VideoInit: TVAspectRatio failed : %d\n",__LINE__);	
			return ret_val;
		}
	}

	// Configure genlock initially (only after TV is initialized)
	if (VidAttr.FullScreenSource != VL_VIDEO_SOURCE_MEMORY) {
		gfx_set_genlock_delay(VidTimings.LineTime);
		gfx_set_genlock_enable(GENLOCK_SYNC_TO_TVENC | GENLOCK_CONTINUOUS);
	}

	// Initialize variables
	using_memory_temporarily = FALSE;
	change_drift_mode = TRUE;           // Force first time call to VidSetDirectVIP

	// Start watchdog timer which handles periodic video actions
	/* <sojan> Modified so that the timer will not be enabled if the
	   IRQ is Zero. This gives the user an option to disable the timer
	   interrupt
	 */  
	if ( VidAttr.TimerIRQ )  {
		if (timer_start() != VL_STATUS_OK ) {
			printk("Error starting timer: %d\n", __LINE__);
			return VL_STATUS_ERROR;
		}
	} else {
		printk("Not enabling timer\n");	
	}

	// Start SAA7114 signal status monitoring
	if (VidAttr.SignalMonitorEnable)
		if (SignalDetectionInit(VidAttr.SignalMonitorGPIONum) != VL_STATUS_OK) {
			printk("SignalDetectionInit failed: %d\n",__LINE__);	
			return VL_STATUS_ERROR;
		}

	if (VidAttr.CRTEnable)
		gfx_set_crt_enable(1);
	else
		gfx_set_crt_enable(0);

	if (VidAttr.TVStandard == VL_TV_STANDARD_NONE)
		gfx_set_tv_enable(0);
	else
		gfx_set_tv_enable(1);
		
	VideoInitialized = TRUE;		

	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Shuts down the video:
//   * Disables video in the video module.
//   * Disables all video-related interrupts.
//	 * Releases all internally-allocated resources.
//
// Parameters:
//   None.
//
// Return Value:
//   VL_STATUS_OK.
//
// Side Effects:
//   None of the video functions may be called after calling this function,
//   except for VL_VideoInit().
//---------------------------------------------------------------------
VL_STATUS VL_VideoEnd()
{

	if (!VideoInitialized) return VL_STATUS_ERROR;

	// Disable video
	VL_VideoEnable(FALSE);

	// Stop watchdog timer
	if(VidAttr.TimerIRQ) /* <sojan> added this to  */
		timer_stop();

	// Stop SAA7114 signal status monitoring
	if (VidAttr.SignalMonitorEnable)
		SignalDetectionEnd();

	gfx_reset_timing_lock();
	
	VideoInitialized = FALSE;

    return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Returns the version of the video library.
//
// Parameters:
//   None.
//
// Return Value:
//   Pointer to a string containing the version of the video library.
//---------------------------------------------------------------------
char * VL_Version()
{
	int i = 0;
	char ch;

	// Skip over the prefix (ProjectRevision) and copy the version number
	// from there
	while ((ch = vl_version[i + 18]) != ' ')
		VidLibVer[i++] = ch;
	VidLibVer[i] = 0;

	return VidLibVer;
}


/**************************************************************************/
/**************************   Video controls   ****************************/
/**************************************************************************/


//---------------------------------------------------------------------
// Description:
//   Enables or disables the video.
//
// Parameters:
//   OnOff - On to enable video, Off to disable.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VideoEnable(boolean OnOff)
{
	VideoSourceType cur_vid_source, new_vid_source;
	int loop = 0;

	// CHECK IF ALREADY THE RIGHT VALUE
	if (gfx_get_video_enable() == OnOff)
		return VL_STATUS_OK;

	// ENABLE OR DISABLE VIDEO
	if (OnOff) {
		if (!VidDisplay.paused) gfx_set_vip_capture_run_mode(VIP_CAPTURE_START_FIELD);
		// ENABLE VIP (IF DVIP IS USED, VIP WILL BE ENABLED WHEN SWITCHING FROM DVIP TO VIP)
		if (VidDisplay.current_video_source == VL_VIDEO_SOURCE_MEMORY) gfx_set_vip_enable(1);
		// ENABLE VIDEO
		gfx_set_video_enable(1);

		/* WORKAROUND FOR HARDWARE PROBLEM */
		/* Need to verify video source register write succeeded */

		if (VidDisplay.current_video_source == VL_VIDEO_SOURCE_MEMORY)
			new_vid_source = VIDEO_SOURCE_MEMORY;
		else
			new_vid_source = VIDEO_SOURCE_DVIP;

		while (1) {
			if (gfx_get_video_source(&cur_vid_source) != GFX_STATUS_OK)
				return VL_STATUS_ERROR;
			if (cur_vid_source == new_vid_source)
				return VL_STATUS_OK;
			gfx_set_video_source(new_vid_source);
			gfx_delay_milliseconds(1);
			if (++loop >= 1000)
				return VL_STATUS_ERROR;
		}

	} else {
		// DISABLE VIDEO
		gfx_set_video_enable(0);
		// To save memory bandwidth when video is disabled
		gfx_set_vip_capture_run_mode(VIP_CAPTURE_STOP_FIELD);
	}
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Checks whether the video is enabled.
//
// Parameters:
//   None.
//
// Return Value:
//    TRUE if video is currently enabled, FALSE otherwise.
//---------------------------------------------------------------------
boolean VL_VideoIsEnabled()
{
    return gfx_get_video_enable();
}    

//---------------------------------------------------------------------
// Description:
//   Pauses the video capture.
//
// Parameters:
//   None.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VideoPause()
{
	if (VidDisplay.paused) return VL_STATUS_OK;
	VidDisplay.paused = TRUE;
	
	if ((VidDisplay.current_video_source == VL_VIDEO_SOURCE_DVIP) ||
		(VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP_ONLY)) {
		// First start capture to memory so current image will be in memory
		gfx_set_vip_enable(1);
		// Finish capturing the current field, and wait for 2 more complete
		// fields to be captured - delay for a total of 3 fields' duration
		gfx_delay_milliseconds(3 * 1000 / VidTimings.RefreshRate);
		// Force switch to VIP
		VidSetDirectVIP(0, 0xC3);
	}
	// Stop the capture so the image will freeze
	gfx_set_vip_capture_run_mode(VIP_CAPTURE_STOP_FIELD);
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Checks whether the video is paused.
//
// Parameters:
//   None.
//
// Return Value:
//    TRUE if video is currently paused, FALSE otherwise.
//---------------------------------------------------------------------
boolean VL_VideoIsPaused()
{
	return VidDisplay.paused;
}

//---------------------------------------------------------------------
// Description:
//   Resumes the video capture.
//
// Parameters:
//   None.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VideoResume()
{
	VidDisplay.paused = FALSE;
	// Restore mode from before pause
	gfx_set_vip_capture_run_mode(VIP_CAPTURE_START_FIELD);
    return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Finds the physical address of row "j" in the video buffer.
//
// Parameters:
//   j - the row of the video buffer to which a pointer is requested
//       (in progressive numbering, starting from line 0).
//
// Return Value:
//   A pointer to row "j" of the video buffer.
//---------------------------------------------------------------------
unsigned short * VL_VideoGetRow(int j)
{
	// In interlaced VIP modes (Bob and Interlaced), the video is in two separate buffers
	if (VidDisplay.current_deinterlacing_method != VL_VIDEO_DEINTERLACING_INTERLEAVE) {
		// In non-interleaved modes the odd and even buffers are separate
		// Multiply VidCaptureBuf.width by 2 to get number of bytes in video line
		return (unsigned short *)(VidCaptureBuf.address + 
			                     ((j & 0x1) * VidCaptureBuf.height/2 + j/2) * VidCaptureBuf.width * 2);
	} else {
		// Multiply VidCaptureBuf.width by 2 to get number of bytes in video line
		return (unsigned short *)(VidCaptureBuf.address + j * VidCaptureBuf.width * 2);
	}
}

//---------------------------------------------------------------------
// Description:
//   Sets the deinterlacing method source for windows of the specified
//   size.
//   This setting will take affect upon the next call to a video
//   window setting function (VL_VideoCfgDisplayWindow() or
//   VL_VideoSetWindow()) for a window of this size.
//
// Parameters:
//   method - the selected deinterlacing method.
//   size - the window size for which the deinterlacing method is to be
//          used.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VideoSetDeinterlacingMethod(
   VL_VideoWindowSize size,
   VL_VideoDeinterlacingMethod method)
{
	if (VidAttr.CRTEnable && (method == VL_VIDEO_DEINTERLACING_NONE))
		return VL_STATUS_ERROR;
		
	switch (size)
	{
		case VL_VIDEO_WINDOW_FULL:
			if (((VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP) ||
				 (VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP_ONLY)) &&
				(method == VL_VIDEO_DEINTERLACING_INTERLEAVE))
				return VL_STATUS_ERROR;
			VidDisplay.full_screen_deinterlacing = method;
			break;
		case VL_VIDEO_WINDOW_SCALED:
			VidDisplay.scaled_video_deinterlacing = method;
			break;
		default:
			return VL_STATUS_BAD_PARAMETER;
	}
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Sets the TV flicker filter type.
//   Note: the TV flicker filter is configured by the Video Library as
//   needed. This method is provided for the use of applications to
//   enable specific "fine tuning".
//
// Parameters:
//   FlickerFilter - the requested flicker filter type.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_TVSetFlickerFilter(VL_TVFlickerFilterType FlickerFilter)
{
	if (VidAttr.TVStandard == VL_TV_STANDARD_NONE)
		return VL_STATUS_OK;
		
	switch (FlickerFilter)
	{
		case VL_TV_FLICKER_FILTER_NONE:
			gfx_set_tv_flicker_filter(TV_FLICKER_FILTER_NONE);
			break;
		case VL_TV_FLICKER_FILTER_INTERLEAVED:
			gfx_set_tv_flicker_filter(TV_FLICKER_FILTER_NORMAL);
			break;
		case VL_TV_FLICKER_FILTER_INTERLACED:
			gfx_set_tv_flicker_filter(TV_FLICKER_FILTER_INTERLACED);
			break;
		default:
			return VL_STATUS_BAD_PARAMETER;
	}
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Sets the preferred source for full screen video.
//   This setting will take affect upon the next call to a video
//   window setting function (VL_VideoCfgDisplayWindow() or
//   VL_VideoSetWindow()) with full screen parameters. 
//   Note: the actual source might be different because the weak signal
//   handling algorithm might determine that DVIP cannot be used at
//   certain times.
//
// Parameters:
//   type - preferred source for full screen video.
//   genlock_delay - the genlock delay to be used, in 27 MHz units.
//                   Relevant only for modes that use DVIP.       
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VideoSetFullScreenSource(VL_VideoSourceType type, unsigned long genlock_delay)
{
	// Check for invalid combinations
	if (((type == VL_VIDEO_SOURCE_DVIP) || (type == VL_VIDEO_SOURCE_DVIP_ONLY)) && 
		((VidDisplay.full_screen_deinterlacing == VL_VIDEO_DEINTERLACING_INTERLEAVE) || VidAttr.CRTEnable))
		return VL_STATUS_ERROR;
		
	VidDisplay.full_screen_source = type;
	VidAttr.GenlockDelay = genlock_delay;
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Sets the hardware to use internal VSYNC signals for genlock.
//   Relevant only when video source is VL_VIDEO_SOURCE_DVIP_ONLY.
//   Whenever the device providing the input video is shut down for any
//   reason, call this function with FALSE to notify to the Video Library
//   that the external VIP signal cannot be used for genlock.
//   When the device is turned on again, call this function with TRUE.
//
// Parameters:
//   OnOff - TRUE if video source is active, FALSE otherwise.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VideoSourceActive(boolean OnOff)
{
	// Check for invalid combinations
	if (VidDisplay.current_video_source != VL_VIDEO_SOURCE_DVIP_ONLY)
		return VL_STATUS_OK;
	if (OnOff) {
		// Reset TV encoder once every odd field (default setting)
		gfx_set_video_source(VIDEO_SOURCE_DVIP);
		gfx_set_tvenc_reset_interval(TVENC_RESET_EVERY_ODD_FIELD);
		gfx_set_genlock_delay(VidAttr.GenlockDelay);
		gfx_set_genlock_enable(GENLOCK_CONTINUOUS);
	} else {		
		// Prevent drift from happening in spite of genlock delay
		gfx_set_video_source(VIDEO_SOURCE_MEMORY);
		gfx_set_tvenc_reset_interval(TVENC_RESET_NEXT_ODD_FIELD);
		gfx_set_genlock_delay(VidTimings.LineTime);
		gfx_set_genlock_enable(GENLOCK_SYNC_TO_TVENC | GENLOCK_CONTINUOUS);
	}		
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Determines whether or not the field polarity of incoming video will
//   be treated as inverted by the Video Library.
//   For Direct VIP modes, this setting will take effect immediately.
//   For video through memory, it will take affect upon the next call
//   to a video window setting function (VL_VideoCfgCaptureBuf() or
//   VL_VideoSetWindow()) for a window of this size.
//
// Parameters:
//   OnOff - TRUE if odd fields should be treated as even fields and vice
//           versa, FALSE otherwise.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_FieldPolarityInvert(boolean OnOff)
{
	VidAttr.FieldPolarityInvert = OnOff;
	gfx_set_top_line_in_odd((int)OnOff);
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Handles maintaining correct graphics aspect ratio on TV or configuring
//   pure video mode 704 (no scaling).
//   Achieving the required aspect ratio involves configuration of the
//   display controller, video overlay hardware and TV out.
//   When only video is displayed set "tv_aspect_ratio" to
//   VL_TV_ASPECT_RATIO_VIDEO_ONLY to get the best video quality; no aspect
//   ratio scaling will be performed.
//   When video and graphics are displayed set "tv_aspect_ratio" to
//   VL_TV_ASPECT_RATIO_GRAPHICS_AND_VIDEO or  VL_TV_ASPECT_RATIO_GRAPHICS_ONLY:
//   - Video is scaled with the aspect ratio's reciprocal, if "tv_aspect_ratio" is
//     VL_TV_ASPECT_RATIO_GRAPHICS_AND_VIDEO.
//   - Video and graphics are blended.
//   - Blended image is scaled to maintain correct graphics aspect ratio.
//   The aspect ratio values are :
//
//            Universal      Carmel Rev. B      Carmel Rev. C
//   - NTSC     11/10            9/8                64/58
//   - PAL      54/59           11/12               64/70
//
// Parameters:
//   tv_aspect_ratio - 
//                 VL_TV_ASPECT_RATIO_GRAPHICS_ONLY:
//                     Both video and graphics are scaled by the required aspect
//                     ratio for graphics on TV according to the TV standard (PAL/NTSC).
//                     To keep correct video aspect ratio on TV, the application
//                     should prescale the video by the reciprocal of the
//                     aspect ratio scaling.
//                 VL_TV_ASPECT_RATIO_GRAPHICS_AND_VIDEO:
//                     Both video and graphics are scaled by the required aspect
//                     ratio for graphics on TV according to the TV standard(PAL/NTSC).
//                     Video is pre-scaled to compensate for the affect of the above
//                     mentioned scaling on the video.
//                 VL_TV_ASPECT_RATIO_VIDEO_ONLY:
//                     No scaling is done. Only video will appear correctly on TV.
//                     Note: in this mode, video aspect ratio on CRT will not be
//                     correct !
//   Members of gfx_struct:
//   GfxAspectRatio  - The factor by which graphics are multiplied to achieve
//                     correct aspect ratio on TV.
//   GfxWidth        - Pointer to variable which the function will set to the
//                     width (in pixels) of the graphics frame buffer after configuring
//                     the video per the requested aspect ratio.
//   GfxHeight       - Pointer to variables which the function will set to the
//                     height (in lines) of the graphics frame buffer after configuring
//                     the video per the requested aspect ratio.
//   GfxPitch        - Pointer to variable which the function will set to the
//                     pitch (in bytes) of the graphics frame buffer after configuring
//                     the video per the requested aspect ratio.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//
// Side Effects:
//   When this function returns, the video is disabled. To re-enable:
//   call the video window setting function(s), alpha window setting
//   function(s) (optional) and video enable function.
//---------------------------------------------------------------------
VL_STATUS VL_TVAspectRatio(VL_TVAspectRatioType tv_aspect_ratio,
						   VL_GfxStruct * gfx_struct)
{
	// CHECK LIMITATION
	// In PAL video prescale must be done in chip, because video line
	// input size to Video Processor is limited to 720 pixels, and
	// prescaling outside the chip would necessitate an input video line
	// of 768 pixels

	if ((tv_aspect_ratio == VL_TV_ASPECT_RATIO_GRAPHICS_ONLY) &&
		(VidAttr.TVStandard == VL_TV_STANDARD_PAL))
		return VL_STATUS_UNSUPPORTED;
	
	// SAVE NEW VALUE
	
	VidAttr.TVAspectRatio = tv_aspect_ratio;

	// CONFIGURE DISPLAY TIMINGS
	
	if (tv_aspect_ratio == VL_TV_ASPECT_RATIO_VIDEO_ONLY)
		VidDisplay.graphics_scale = FALSE;
	else
		VidDisplay.graphics_scale = TRUE;			
	VidSetTVTimings();
	
	// RESET VIDEO OVERLAY SCALING
	
	gfx_set_video_downscale_enable(0);
	gfx_set_video_upscale(VidTimings.Width, 0, VidTimings.Width, 0);

	// CONFIGURE VIDEO OVERLAY SCALING
	
	if (tv_aspect_ratio == VL_TV_ASPECT_RATIO_VIDEO_ONLY)
	{
		// No need to prescale video
		VidDisplay.horzFactor.num = VidDisplay.horzFactor.denom = 1;
	} else {
		// Determine video scale factor needed to initially fit the graphics resolution
		if (VidAttr.TVStandard == VL_TV_STANDARD_NTSC) {
			// Set the horizontal scale factor for 704->640 (needed by window functions)
			if (gfx_chip_revision <= SC1200_REV_B3) {
				VidDisplay.horzFactor.num = 8;
				VidDisplay.horzFactor.denom = 9;
				if (tv_aspect_ratio == VL_TV_ASPECT_RATIO_GRAPHICS_AND_VIDEO) {
					// Downscale done in chip
					gfx_set_video_downscale_config(VIDEO_DOWNSCALE_DROP_1_OF, 9);
					gfx_set_video_downscale_coefficients(0x0, 0x0, 0x4, 0xC);
					gfx_set_video_downscale_enable(1);
					gfx_set_video_upscale(VidTimings.Width, 0, VidTimings.Width, 0);
				}					
			} else {
				VidDisplay.horzFactor.num = 58;
				VidDisplay.horzFactor.denom = 64;
				if (tv_aspect_ratio == VL_TV_ASPECT_RATIO_GRAPHICS_AND_VIDEO) {
					// Downscale done in chip
					// Use combination of upscale and downscale to achieve required downscale factor
					gfx_set_video_downscale_config(VIDEO_DOWNSCALE_DROP_1_OF, 2);
					gfx_set_video_downscale_coefficients(0x0, 0x0, 0x4, 0xC);
					gfx_set_video_downscale_enable(1);
					gfx_set_video_upscale(VidDisplay.horzFactor.denom, 0, VidDisplay.horzFactor.num * 2, 0);
				}					
			}				
		} else if (VidAttr.TVStandard == VL_TV_STANDARD_PAL) {
			// Set the horizontal scale factor for 704->768 (needed by window functions)
			if (gfx_chip_revision <= SC1200_REV_B3) {
				VidDisplay.horzFactor.num = 12;
				VidDisplay.horzFactor.denom = 11;
				// Upscale must be done in chip for PAL
				gfx_set_video_upscale(VidDisplay.horzFactor.denom, 0, VidDisplay.horzFactor.num, 0);
				gfx_set_video_downscale_enable(0);
			} else {
				VidDisplay.horzFactor.num = 70;
				VidDisplay.horzFactor.denom = 64;
				// Upscale must be done in chip for PAL
				gfx_set_video_upscale(VidDisplay.horzFactor.denom, 0, VidDisplay.horzFactor.num, 0);
				gfx_set_video_downscale_enable(0);
			}				
		}			
	}		

	// CONFIGURE TV OUT
	
	switch (VidAttr.TVStandard) {
		case VL_TV_STANDARD_NTSC:
			if (tv_aspect_ratio == VL_TV_ASPECT_RATIO_VIDEO_ONLY)
				gfx_set_tv_format(TV_STANDARD_NTSC, GFX_ON_TV_NO_SCALING);
			else
				gfx_set_tv_format(TV_STANDARD_NTSC, GFX_ON_TV_SQUARE_PIXELS);
			break;
		case VL_TV_STANDARD_PAL:
			if (tv_aspect_ratio == VL_TV_ASPECT_RATIO_VIDEO_ONLY)
				gfx_set_tv_format(TV_STANDARD_PAL, GFX_ON_TV_NO_SCALING);										
			else
				gfx_set_tv_format(TV_STANDARD_PAL, GFX_ON_TV_SQUARE_PIXELS);
			break;
		case VL_TV_STANDARD_NONE:
			return VL_STATUS_ERROR;  // No need to keep aspect ratio when no TV
		default:
			return VL_STATUS_BAD_PARAMETER;
	}

	// SET RETURN VALUES
	
	gfx_struct->GfxWidth = gfx_get_hactive();
	gfx_struct->GfxHeight = gfx_get_vactive();
	gfx_struct->GfxPitch = gfx_get_display_pitch();
	
	// Graphics aspect ratio is the reciprocal of the video prescaling factor
	gfx_struct->GfxAspectRatio.num = VidDisplay.horzFactor.denom;
	gfx_struct->GfxAspectRatio.denom = VidDisplay.horzFactor.num;
	
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Sets the Display Controller timing for the current TV standard,
//   taking into account whether graphics aspect ratio is required to
//   be kept on TV, and whether desired full screen source is DVIP.
//
// Parameters:
//   None.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VidSetTVTimings(void)
{
	VL_GfxTVResolution GfxTVResolution;
	DISPLAYMODE Tim;
	int comp_enable;
	unsigned short gfx_bpp;
	/* <sojan> Added code to query current compression settings
	   and restore later
	 */  
	unsigned short compression_size;
	unsigned short compression_pitch;
	unsigned long  compression_offset;

	// Configure TV output

	switch (VidAttr.TVStandard) {
		case VL_TV_STANDARD_NTSC:
			if (VidDisplay.graphics_scale)
				GfxTVResolution = VL_GFX_TV_RESOLUTION_NTSC;
			else
				GfxTVResolution = VL_GFX_TV_RESOLUTION_NTSC_NO_SCALING;
			break;
		case VL_TV_STANDARD_PAL:
			if (VidDisplay.graphics_scale)
				GfxTVResolution = VL_GFX_TV_RESOLUTION_PAL;
			else
				GfxTVResolution = VL_GFX_TV_RESOLUTION_PAL_NO_SCALING;
			break;
		case VL_TV_STANDARD_NONE:
			return VL_STATUS_UNSUPPORTED;  // Cannot set timings for unknown TV format
		default:
			return VL_STATUS_BAD_PARAMETER;
	}
	// CONFIGURE DISPLAY CONTROLLER

	// Save graphics compression state

	comp_enable = gfx_get_compression_enable();
	gfx_bpp = gfx_get_display_bpp();
        compression_size = gfx_get_compression_size();
	compression_pitch = gfx_get_compression_pitch();
	compression_offset = gfx_get_compression_offset();

	Tim = GfxDisplayTimings[GfxTVResolution];

	if ((VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP) ||
		(VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP_ONLY))
	// Modify graphics timings for genlock mode - needed for DVIP
		gfx_set_display_timings(gfx_bpp, Tim.flags,
			Tim.hactive, Tim.hblankstart, Tim.hsyncstart,
			Tim.hsyncend, Tim.hblankend, Tim.htotal,
			Tim.vactive, Tim.vblankstart, Tim.vsyncstart - 8,
			Tim.vsyncend - 8, Tim.vtotal - 9, Tim.vtotal - 9,
			Tim.frequency);
	else
		gfx_set_display_timings(gfx_bpp, Tim.flags,
			Tim.hactive, Tim.hblankstart, Tim.hsyncstart,
			Tim.hsyncend, Tim.hblankend, Tim.htotal,
			Tim.vactive, Tim.vblankstart, Tim.vsyncstart,
			Tim.vsyncend, Tim.vblankend, Tim.vtotal,
			Tim.frequency);

	// Restore graphics compression state and set compression parameters

	if (comp_enable) {
		gfx_set_compression_size(compression_size);
		gfx_set_compression_pitch(compression_pitch);
		gfx_set_compression_offset(compression_offset);
		gfx_set_compression_enable(1);
	}

	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Sets video processor output signal to VOP or TFT.
//
// Parameters:
//   SignalType - the requested output signal type.
//
// Return Value:
//   VL_STATUS_OK if successful, otherwise one of the error return values.
//---------------------------------------------------------------------
VL_STATUS VL_VOPSignalSelect(VL_VOPSignalSelectType SignalType)
{
	unsigned long pmr;

	switch (SignalType)
	{
		case VL_SELECT_VOP_SIGNALS:
			// Select VOP signals : PMR[15] = 1
			pmr = OS_IoReadDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR);
			pmr |= 0x8000;
			OS_IoWriteDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR, pmr);
			// Enable VOP signals : PMR[23] = 1
			pmr = OS_IoReadDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR);
			pmr |= 0x800000;
			OS_IoWriteDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR, pmr);
			break;
		case VL_SELECT_TFT_SIGNALS:
			// Select TFT signals : PMR[15] = 0
			pmr = OS_IoReadDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR);
			pmr &= ~0x8000;
			OS_IoWriteDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR, pmr);
			// Enable TFT signals : PMR[23] = 1
			pmr = OS_IoReadDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR);
			pmr |= 0x800000;
			OS_IoWriteDWord(SC1200_CB_BASE_ADDR + SC1200_CB_PMR, pmr);
			break;
		default:
			return VL_STATUS_BAD_PARAMETER;
	}
	return VL_STATUS_OK;
}
