/* NSC_LIC_BSD
 *
 * National Semiconductor Corporation Open Source License for
 *
 *    Video library for Geode SCx2x0
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) 2000-2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use by Licensee of this Software in source and
 * binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must either:
 *     (a) reproduce the above copyright notice, this list of conditions
 *         and the following disclaimer in the documentation and/or
 *         other materials provided with the distribution, or
 *     (b) if Licensee elects to redistribute any Software not in
 *         accordance with (a), then Licensee shall indemnify National
 *         Semiconductor Corporation and all copyright holders and
 *         contributors for any claims related to such Software (and pay
 *         all costs and damages, including all attorney fees)
 *
 *   * Neither the name of the National Semiconductor Corporation nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF
 * YOUR JURISDICTION. It is Licensee's responsibility to comply with
 * any export regulations applicable in Licensee's jurisdiction. Under
 * CURRENT (July 2001) U.S. export regulations this software
 * is eligible for export from the U.S. and can be downloaded by or
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran,
 * Syria, Sudan, Afghanistan and any other country to which the U.S.
 * has embargoed goods and services.
 *
 * END_NSC_LIC_BSD */
 
/*--------------------------------------------------------------------------
 * Copyright (c) 2000-2001 by National Semiconductor Corporation
 * All rights reserved.
 *--------------------------------------------------------------------------
 * File Contents:
 *           Functions for configuring video capture and display.
 * Project : Video library for Geode SCx2x0
 * Created : November, 2000
 * Author  : Galit Heller & Eyal Cohen
 *--------------------------------------------------------------------------
 * $Date: 8/10/01 6:58a $
 * $RCSfile: vl_win.c $
 * $ProjectRevision: 3.1 $
 * $Revision: 6 $
 *------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*-------------------------   Include files   ----------------------------*/
/*------------------------------------------------------------------------*/
#include "gfx_rtns.h"
#include "gfx_defs.h"
#include "vl_rtns.h"
#include "vl_defs.h"
#include "vl_os.h"

/*------------------------------------------------------------------------*/
/*----------------   Constants and macros definitions  -------------------*/
/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*------------   Data Types definitions and declarations   ---------------*/
/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/*-------------   Variable definitions and declarations   ----------------*/
/*------------------------------------------------------------------------*/
boolean using_memory_temporarily = 0;
boolean change_drift_mode = 0;
unsigned long vip_even_base = 0, vip_odd_base = 0;
VL_MemBuf VidCaptureBuf;
VL_MemBuf VidDisplayBuf;

/*------------------------------------------------------------------------*/
/*-----------------   Private function declarations   --------------------*/
/*------------------------------------------------------------------------*/
void UpdateVipType(void);
void VidSetTearingCorrection(void);

/*-------------*/
/*-- Externs --*/
/*-------------*/
extern VidTimingsChar VidTimings;
extern int field_count;

/**************************************************************************/
/******************   Video window (scale/crop/zoom)   ********************/
/**************************************************************************/

/*
	One hardware video overlay window is available for use by the
	following functions.

	Use the following sequence to change video window size/location :
	1. Verify that video is not paused if you intend to change
	   video decoder settings.
	2. Set the video decoder input window using calls to :
		   gfx_set_decoder_input_offset(x , y);
		   gfx_set_decoder_input_size(w, h);
	3. Set the video decoder output window and scale the active video to the
	   requested size using calls to :
		   gfx_set_decoder_output_size(w, h);
		   gfx_set_decoder_scale(srcWidth, srcHeight, dstWidth, dstHeight);
	4. Call VL_VideoSetWindow() to set the video window position and size.

	The sequence described above detailed the complete steps needed in
	order to change the video window size and location.
	Depending on the requested change of the video window, only part
	of the sequence may need to be executed. This can be done using
	specific video window functions - see more in this file. 
*/

//---------------------------------------------------------------------
// Description:
//   Sets the video window position and size, without scaling the video
//   (except for the aspect ratio scaling).
//   Sets entire path for video window programming:
//     Selects between VIP and DVIP.
//     Configures the video capture to memory (if VIP).
//     Configures the getting the video from memory (if VIP).
//     Configures the displayed window on the screen.
//
// Parameters:
//   mem_buf - Address of video buffer to be used (not needed for DVIP).
//             Should be large enough to contain two fields (odd/even).
//             This address should be 16-byte aligned. If not, the actual
//             address used will be this address truncated down to the
//             closest 16 byte aligned address.
//             Note: this argument is not relevant when setting a full screen
//             video window in DVIP ONLY mode
//   x       - Horizontal position where left-most pixel of window should be placed.
//             If graphics aspect ratio is kept, the actual position is aligned
//             to the nearest 8/12 (NTSC/PAL) pixel multiple.
//             If x is negative odd, it is rounded up to an even number (i.e. one
//             is added).
//   y       - Vertical position where top line of window should be placed.
//   width   - Horizontal number of pixels in the window.
//             Note: mem_buf.width is rounded down to the nearest 8-pixel multiple, 
//             so width should be less than or equal to this rounded value.
//   height  - Vertical number of lines in the window.
//
// Return Value:
//   VL_STATUS_OK.
//
// Side Effects:
//   This function sets the width field of the mem_buf structure
//   to the actual capture buffer width according to HW limitations (rounded
//   down to the nearest 8-pixel multiple).
//---------------------------------------------------------------------
VL_STATUS VL_VideoSetWindow(VL_MemBuf * mem_buf,
   short x, short y, unsigned short width, unsigned short height)
{
	short display_x, display_y, display_width_sub, display_height_sub;
	
	// CAN'T CHANGE THE WINDOW SIZE IF VIDEO IS PAUSED SINCE VIP HARDWARE
	// CHANGES CANNOT AFFECT THE IMAGE WHICH IS ALREADY CAPTURED IN MEMORY.
	if (VidDisplay.paused)
		return VL_STATUS_ERROR;

	// Line pitch in memory must be an 8 pixel multiple, due to VIP hardware limitation
	// of 16 byte aligned line base addresses
	// This affects the memory buffer used for both capture to memory and 
	// display from memory, and the video window size
	mem_buf->width = mem_buf->width & ~0x7;
	
	// Adjust window width according to memory buffer width
	if (VidAttr.TVAspectRatio == VL_TV_ASPECT_RATIO_GRAPHICS_ONLY) {
		// Video scaling is done outside chip, so video buffer contains width after scaling
		if (ASPECT_RATIO_VIDEO_PRESCALE(width) > mem_buf->width)
			width = mem_buf->width * VidDisplay.horzFactor.denom / VidDisplay.horzFactor.num;
	} else {
		// Video scaling is done inside chip or not at all, so video buffer must contain video before
		// scaling
		if (width > mem_buf->width)
			width = mem_buf->width;
	}				

	// SET OVERLAY SCREEN POSITION AND SIZE, TAKING INTO ACCOUNT NEGATIVE OFFSETS

	if (x < 0) {
		display_x = 0;
		display_width_sub = -x;
	} else {
		display_x = x;
		display_width_sub = 0;
	}
	if (y < 0) {
		display_y = 0;
		display_height_sub = -y;
	} else {
		display_y = y;
		display_height_sub = 0;
	}
	
	if (display_width_sub > width) display_width_sub = width;
	if (display_height_sub > height) display_height_sub = height;
	
	VL_VideoCfgDisplayWindow(display_x, display_y, width - display_width_sub, height - display_height_sub);

	// CONFIGURE CAPTURE TO MEMORY (VIP MODULE) AND DISPLAY FROM MEMORY HARDWARE (GXM)
	// THIS IS DONE EVEN IF DVIP IS ACTIVE, IN ORDER TO FACILITATE TRANSITION TO VIP
	// WHEN NEEDED

	VL_VideoCfgCaptureBuf(mem_buf);
	
	// Positive x, y values do not affect the memory window, because they relate to the
	// where the video is displayed on screen.
	if (x >= 0) x = 0;
	if (y >= 0) y = 0;
	// Negative x, y values mean that the memory window should start at non-zero offsets
	VL_VideoCfgMemoryWindow(mem_buf, (unsigned short)(-x), (unsigned short)(-y), width - display_width_sub, height - display_height_sub);

    return VL_STATUS_OK;
}

/*
	Following are the component functions of the video path configuration,
	each of which may be called separately. It is the responsibility
	of the caller that individual calls have "logical" effects.
	These functions provide more flexibility to enable manipulation such
	as window cropping, creating video album etc.
	Note that all these functions are called from the main function
	VL_VideoSetWindow().
*/

//---------------------------------------------------------------------
// Description:
//   Sets the area in memory to which the VIP will capture,
//   effective from the beginning of the next field, by setting
//   VIP bases and pitch registers according to deinterlacing method.
//
// Parameters:
//   mem_buf - Address of the buffer to which the VIP should capture.
//             The width field of mem_buf should be a multiple of 8 pixels.
//             This address should be 16-byte aligned. If not, the actual
//             address used will be this address truncated down to the
//             closest 16 byte aligned address.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VideoCfgCaptureBuf(VL_MemBuf * mem_buf)
{
	unsigned short vip_line, odd_height;
	
	// Nothing to do for DVIP ONLY mode because memory is never used
	// for full screen
	if (VidDisplay.current_video_source == VL_VIDEO_SOURCE_DVIP_ONLY)
		return VL_STATUS_OK;

	// Align capture buffer address to 16 bytes due to capture hardware
	// restriction
	mem_buf->address &= ~(unsigned long)0xf;
	
	// Save capture buffer parameters
	VidCaptureBuf = *mem_buf;
	
	vip_line = VidCaptureBuf.width << 1;  // vip_line is in bytes and buffer width in (2-byte) pixels
	
	vip_odd_base = VidCaptureBuf.address - (unsigned long)gfx_phys_fbptr;
	
	// Configure VIP capture location according to deinterlacing method
		
	if (VidDisplay.current_deinterlacing_method != VL_VIDEO_DEINTERLACING_INTERLEAVE)
	{			 
		// UPDATE VIP WINDOW FOR NEW SIZE
		// Changing addresses of data going into memory from VIP to match 
		// the destination size.

		// For odd number of lines, odd field contains one line more than the even field									
		if (VidCaptureBuf.height & 0x1)
			odd_height = (VidCaptureBuf.height + 1)/2;
		else
			odd_height = VidCaptureBuf.height/2;					
		vip_even_base = vip_odd_base + vip_line * odd_height;
		gfx_set_vip_pitch(vip_line);
	} else {
		vip_even_base = vip_odd_base + vip_line;
		gfx_set_vip_pitch(vip_line << 1);
	}
	if ((VidAttr.FieldPolarityInvert) &&
	    (VidDisplay.current_deinterlacing_method == VL_VIDEO_DEINTERLACING_INTERLEAVE))
		gfx_set_vip_base(vip_odd_base, vip_even_base);
	else				
		gfx_set_vip_base(vip_even_base, vip_odd_base);

	gfx_set_vip_last_line((VidCaptureBuf.height + 1) >> 1);

	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Check whether VIP has already started capturing video into the
//   buffer last set by the call to VL_VideoCfgCaptureBuf().
//
// Parameters:
//    None.
//
// Return Value:
//   TRUE -  If VIP has already started capturing video into the buffer.
//   FALSE - Otherwise. 
//---------------------------------------------------------------------
boolean VL_VideoCaptureBufIsUpdated()
{
	// Check whether all video and vbi data base registers are updated
	return (gfx_test_vip_bases_updated() ||
	// If not - check specifically whether video data base registers are updated
			((vip_odd_base + (unsigned long)gfx_phys_fbptr == gfx_get_vip_base(1)) &&
			 (vip_even_base + (unsigned long)gfx_phys_fbptr == gfx_get_vip_base(0))));
}

//---------------------------------------------------------------------
// Description:
//   Selects the sub-window of a memory buffer (which contains captured
//   video data) to be sent to the video processor for processing, by
//   setting GXm video size.
//
// Parameters:
//    mem_buf - Pointer to the buffer in which the video to be shown resides.
//    x - horizontal offset from beginning of the video buffer.
//        If x is negative odd, it is rounded up to an even number (i.e. one
//        is added).
//    y - vertical offset from beginning of the video buffer.
//    width  - number of pixels in each buffer row.
//             If width is odd, it is rounded down to an even number (i.e. one
//              is subtracted).
//    height - number of rows in progressive numbering (i.e. 1..480) even
//             for Interlaced/Bob modes.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VideoCfgMemoryWindow(VL_MemBuf * mem_buf,
   unsigned short x, unsigned short y, unsigned short width, unsigned short height)
{
	unsigned short odd_height;
	VidDisplayBuf = *mem_buf;
	
	// For DVIP ONLY mode because memory is never used for full screen,
	// so only need to configure Video Processor Input Line Size
	// (Input line size must be even due to Video Processor limitation)
	
	if (VidDisplay.current_video_source == VL_VIDEO_SOURCE_DVIP_ONLY) {
		if (VidAttr.TVAspectRatio == VL_TV_ASPECT_RATIO_GRAPHICS_ONLY)
			gfx_set_video_size(ASPECT_RATIO_VIDEO_PRESCALE(VidTimings.Width) & (~0x1), VidTimings.Height);
		else			
			gfx_set_video_size(VidTimings.Width, VidTimings.Height);
		return VL_STATUS_OK;
	}
	
	// Y offset effect is implemented by increasing the video buffer offset in the display
	// controller
	// Y is in progressive numbering, so divide by two for interlaced modes
	
	if (VidDisplay.current_deinterlacing_method != VL_VIDEO_DEINTERLACING_INTERLEAVE)
		y = y/2;
		
	VidDisplay.memory_address_odd = VidDisplayBuf.address + 
									VidDisplayBuf.width * 2 * y;
									
		// For odd number of lines, odd field contains one line more than the even field									
		if (VidDisplayBuf.height & 0x1)
			odd_height = (VidDisplayBuf.height + 1)/2;
		else
			odd_height = VidDisplayBuf.height/2;					
			
	VidDisplay.memory_address_even = VidDisplayBuf.address + 
									 VidDisplayBuf.width * 2 * (odd_height + y);

	// Set the video buffer offset to start at odd field - suitable for interleaved video
	// (and for non-interleaved video, where the odd field will be shown first)

	gfx_set_video_offset(VidDisplay.memory_address_odd - (unsigned long)gfx_phys_fbptr);
	
	// Set video buffer size according to deinterlacing method
	if (VidDisplay.current_deinterlacing_method != VL_VIDEO_DEINTERLACING_INTERLEAVE)				 
		gfx_set_video_size(VidDisplayBuf.width, height >> 1);
	else
		gfx_set_video_size(VidDisplayBuf.width, height);
		
	// X offset effect is implemented by cropping the start of video line in video processor
	// ROUND THE X OFFSET TO 2 PIXEL (1 DWORD) MULTIPLES DUE TO VIDEO PROCESSOR LIMITATION
	x &= (~0x1);	
	gfx_set_video_left_crop(x);				 
	return VL_STATUS_OK;
}

//---------------------------------------------------------------------
// Description:
//   Sets the area of the screen on which the video will be displayed,
//   and handles all aspects relating to the displayed window:
//   - Updates video display variable (window size, deinterlacing mode,
//     tearing correction state).
//   - For full screen - reset TV encoder once every odd field;
//   - For scaled window - reset TV encoder at the next odd field
//   - If switching from scaled window to DVIP, go through VIP to "smooth"
//     the switch.
//   - Call UpdateVipType() to select between VIP/DVIP.
//   - Set video overlay window according to VIP/DVIP and scaling factor.
//   - Set deinterlacing method.
//
// Parameters:
//    x - horizontal offset from beginning of the screen.
//        If graphics aspect ratio is kept, the actual position is aligned
//        to the nearest 8/12 (NTSC/PAL) pixel multiple.
//    y - vertical offset from beginning of the screen.
//    width -  number of pixels horizontally.
//             width is rounded (down) by this function to a 2 pixel multiple
//    height - number of rows in progressive numbering (i.e. 1..480) even
//             for Interlaced/Bob modes.
//
// Return Value:
//   VL_STATUS_OK.
//---------------------------------------------------------------------
VL_STATUS VL_VideoCfgDisplayWindow(unsigned short x, unsigned short y,
   unsigned short width, unsigned short height)
{	
	// SAVE VALUES

	VidDisplay.current_window_x = x;
	VidDisplay.current_window_y = y;
	VidDisplay.current_window_width = width;
	VidDisplay.current_window_height = height;
	
    // UPDATE VIDEO DISPLAY VARIABLES
    if ((x == 0) && (y == 0) && (width == VidTimings.Width) && (height == VidTimings.Height)) {
		// FULL SCREEN
		if ((VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP) &&
			(VidDisplay.current_video_source != VL_VIDEO_SOURCE_DVIP)) {
			// Check whether transitioning from scaled genlock to full screen genlock	
			if ((VidDisplay.current_window_size == VL_VIDEO_WINDOW_SCALED) && gfx_get_genlock_enable())
				change_drift_mode = TRUE;
			// Transition to full screen DVIP - set "using_memory_temporarily" so that:
			// Full screen VIP is activated first (to "smooth" the switch), and
			// tearing correction is disabled immediately
			using_memory_temporarily = TRUE;
		}		
		VidDisplay.current_window_size = VL_VIDEO_WINDOW_FULL;
	} else {
		// SCALED WINDOW
		// Check whether transitioning from full screen genlock w/ drift (VIP)
		// to scaled genlock (w/o drift)
		if ((VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP) &&
			(VidDisplay.current_window_size == VL_VIDEO_WINDOW_FULL) &&
			gfx_get_genlock_enable())
			change_drift_mode = TRUE;
		VidDisplay.current_window_size = VL_VIDEO_WINDOW_SCALED;		
    }

	// PERFORM HARDWARE CONFIGURATION FOR THE NEW DEINTERLACING METHOD AND
	// UPDATE DEINTERLACING METHOD - MUST BE DONE BEFORE CALLING CAPTURE AND MEMORY FUNCTIONS

	VidSetDeinterlacingMethod();

	// Select and configure VIP/DVIP mode
	UpdateVipType();
	
	// SET THE VIDEO OVERLAY POSITION AND SIZE
	
	// IF ASPECT RATIO SCALING IS PERFORMED ON THE VIDEO, ADJUSTMENTS MUST BE MADE:
	// HORIZONTAL START POSITION AND SIZE ARE MULTIPLED BY THE ASPECT RATIO SCALING, AND
	// HORIZONTAL START POSITION IS ROUNDED TO 8 or 12 PIXEL MULTIPLES
	// (TO MINIMIZE SCALING ARTIFACTS)
	if (VidDisplay.horzFactor.num != VidDisplay.horzFactor.denom) {
		x = ASPECT_RATIO_VIDEO_PRESCALE(x);
		width = ASPECT_RATIO_VIDEO_PRESCALE(width);
		if (VidAttr.TVStandard == VL_TV_STANDARD_NTSC)
			x = (x + 4) & (~0x7);
		else if (VidAttr.TVStandard == VL_TV_STANDARD_PAL)
			x = ((x + 6) / 12) * 12;
	}			

	gfx_set_video_window(x, y + VidDisplay.extra_y_offset, width, height);
	
	// SET TEARING CORRECTION STATUS FOR NEW WINDOW
	VidSetTearingCorrection();
	
	return VL_STATUS_OK;
}

/**************************************************************************/
/************************   Private functions   ***************************/
/**************************************************************************/

//---------------------------------------------------------------------------
// Fills the buffer from which video has been displayed with
// "value" pixels.
//---------------------------------------------------------------------------
void clear_video_buffer(unsigned short value)
{
	unsigned long buf_start, buf_end;
	
	if (VidDisplayBuf.address == 0) return;
	
	buf_start = VidDisplayBuf.address - (unsigned long)gfx_phys_fbptr;
	buf_end = buf_start + VidDisplayBuf.width * 2 * VidDisplayBuf.height;

	for (; buf_start < buf_end; buf_start += 2)
		WRITE_FB16(buf_start, value);
}

//---------------------------------------------------------------------------
// Utility routine which determines whether to enable VIP or Direct VIP,
// according to the destination window size, and the input
// and output offsets, and then makes the setting accordingly.
//---------------------------------------------------------------------------
void UpdateVipType(void)
{
	// Check whether to use DVIP or VIP according to situation
	if ((VidDisplay.current_window_size == VL_VIDEO_WINDOW_FULL) &&
	    (VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP) &&
		!VidDisplay.paused &&
		!using_memory_temporarily) {
		// DVIP can be applied
		// Check if we are not already in DVIP
		if (VidDisplay.current_video_source != VL_VIDEO_SOURCE_DVIP) {
			// Change to DVIP
			VidSetDirectVIP(1, 0xD2);
		}												
	} else if ((VidDisplay.current_window_size == VL_VIDEO_WINDOW_FULL) &&
			   (VidDisplay.full_screen_source == VL_VIDEO_SOURCE_DVIP_ONLY)) {
		if (VidDisplay.current_video_source != VL_VIDEO_SOURCE_DVIP_ONLY) {
			// Change to DVIP ONLY
			VidSetDirectVIP(1, 0xD3);
		}												   
	} else {
		// VIP needed
		// Check if transitioning from DVIP to VIP or changing VIP type
		if ((VidDisplay.current_video_source != VL_VIDEO_SOURCE_MEMORY) ||
			((VidDisplay.full_screen_source == VL_VIDEO_SOURCE_MEMORY) && gfx_get_genlock_enable()) ||
			((VidDisplay.full_screen_source != VL_VIDEO_SOURCE_MEMORY) && !gfx_get_genlock_enable()) ||
			change_drift_mode) {
			// Change to VIP
			VidSetDirectVIP(0, 0xC2);
		}	
	}
}

//---------------------------------------------------------------------------
// VidSetDirectVIP
//
// Enables or disables Direct VIP.
// onOff - TRUE, makes the necessary configurations for displaying Direct VIP,
//         including settings needed for displaying Direct VIP on TV.
//         FALSE, undoes the Direct VIP configurations in order to enable
//         display of VIP. VIP is either with or without genlock, according
//         to the requested full screen source.
//---------------------------------------------------------------------------
void VidSetDirectVIP(boolean onOff, int debug)
{
	if (onOff) {
		// If we are not in genlock, then we need to modify timings to genlock mode
		if (!gfx_get_genlock_enable())
			VidSetTVTimings();
		// Special setting for DVIP
		if (VidAttr.TVStandard == VL_TV_STANDARD_NTSC)
			VidDisplay.extra_y_offset = EXTRA_Y_FOR_DVIP_NTSC;
		else
			VidDisplay.extra_y_offset = EXTRA_Y_FOR_DVIP_PAL;					
		gfx_set_video_window(0, VidDisplay.extra_y_offset,
							 ASPECT_RATIO_VIDEO_PRESCALE(VidTimings.Width),
			                 VidTimings.Height);
		// Configure genlock
		gfx_set_genlock_delay(VidAttr.GenlockDelay);
		// Configure TV encoder reset to default value (changed for scaled VIP)			
		gfx_set_tvenc_reset_interval(TVENC_RESET_EVERY_ODD_FIELD);
		// Set video source to DVIP
		gfx_set_video_source(VIDEO_SOURCE_DVIP);
		gfx_set_vip_enable(0);
		clear_video_buffer(0x1080);
		using_memory_temporarily = FALSE;
		VidDisplay.current_video_source = VidDisplay.full_screen_source;
		if (VidDisplay.current_video_source == VL_VIDEO_SOURCE_DVIP_ONLY) {
			gfx_set_genlock_enable(GENLOCK_CONTINUOUS);
		} else {
			gfx_set_genlock_enable(GENLOCK_CONTINUOUS | GENLOCK_TIMEOUT);
		}
#ifdef DEBUG
		OS_IoWriteByte(0x80, debug);
#endif
	} else {
		// Enable VIP capture - needed when transitioning from DVIP
		if (gfx_get_video_enable())
			gfx_set_vip_enable(1);

		// Reset special DVIP setting
		VidDisplay.extra_y_offset = 0;
		// Set video source to memory
		gfx_set_video_source(VIDEO_SOURCE_MEMORY);
		VidDisplay.current_video_source = VL_VIDEO_SOURCE_MEMORY;

		// Transition from one VIP to another

		if (((VidDisplay.full_screen_source == VL_VIDEO_SOURCE_MEMORY) && gfx_get_genlock_enable()) ||
			((VidDisplay.full_screen_source != VL_VIDEO_SOURCE_MEMORY) && !gfx_get_genlock_enable()))
			VidSetTVTimings();

		// Set VIP to either drift/genlock or no drift/no genlock mode

		if (VidDisplay.full_screen_source == VL_VIDEO_SOURCE_MEMORY) {
			using_memory_temporarily = FALSE;
			// Reset TV encoder once every odd field (default setting)
			gfx_set_tvenc_reset_interval(TVENC_RESET_EVERY_ODD_FIELD);
			gfx_set_genlock_enable(0);
		} else {
			if (VidDisplay.current_window_size == VL_VIDEO_WINDOW_FULL) {
				// Create drift
				// Restore VPHASE value
				int vphase = gfx_get_tv_vphase();
				if (vphase > 20)
					gfx_set_tv_vphase(vphase - VPHASE_OFFSET);
				gfx_set_tvenc_reset_interval(TVENC_RESET_EVERY_ODD_FIELD);
			} else {
				// Scaled window - prevent drift from happening in genlock to
				// encoder with the current genlock delay
				gfx_set_tvenc_reset_interval(TVENC_RESET_NEXT_ODD_FIELD);
				using_memory_temporarily = FALSE;
			}
			// Set genlock delay to shorten GXm field time and thus create a
			// drift of GXm VSYNC towards VIP VSYNC so that transition to DVIP
			// will cause a minimal change in GXm VSYNC
			gfx_set_genlock_delay(VidTimings.LineTime);
			gfx_set_genlock_enable(GENLOCK_SYNC_TO_TVENC | GENLOCK_CONTINUOUS);
			change_drift_mode = FALSE;
		}
#ifdef DEBUG
		OS_IoWriteByte(0x80, debug);
#endif
	}
	field_count = 0;
}

//---------------------------------------------------------------------------
// Updates the current deinterlacing method, and performs the appropriate Video
// Processor settings.
//---------------------------------------------------------------------------
void VidSetDeinterlacingMethod()
{
	VL_VideoDeinterlacingMethod new_deinterlacing_method;
	
	// UPDATE DEINTERLACING METHOD VARIABLE
	if (VidDisplay.current_window_size == VL_VIDEO_WINDOW_FULL)
		new_deinterlacing_method = VidDisplay.full_screen_deinterlacing;
	else
		new_deinterlacing_method = VidDisplay.scaled_video_deinterlacing;
	if (new_deinterlacing_method != VidDisplay.current_deinterlacing_method) {
		VidDisplay.current_deinterlacing_method = new_deinterlacing_method;
	
		// PERFORM HARDWARE CONFIGURATION FOR THE NEW DEINTERLACING METHOD
		switch (new_deinterlacing_method)
		{
			case VL_VIDEO_DEINTERLACING_NONE:
				gfx_set_video_interlaced(1);
				gfx_set_video_upscale(0, 1 + 1, 0, 1 + 1); // Do not upscale
				gfx_set_vertical_scaler_offset(0);
				gfx_set_tv_flicker_filter(TV_FLICKER_FILTER_INTERLACED);			
				break;
			case VL_VIDEO_DEINTERLACING_INTERLEAVE:
				gfx_set_video_interlaced(0);
				gfx_set_video_upscale(0, 1 + 1, 0, 1 + 1); // Do not upscale
				gfx_set_vertical_scaler_offset(0);
				gfx_set_tv_flicker_filter(TV_FLICKER_FILTER_NORMAL);
				gfx_set_video_offset(VidDisplayBuf.address - (unsigned long)gfx_phys_fbptr);
				break;
			case VL_VIDEO_DEINTERLACING_BOB:
				gfx_set_video_interlaced(0);
				gfx_set_video_upscale(0, 1 + 1, 0, 2 + 1); // Upscale by 2
				gfx_set_vertical_scaler_offset(1);
				gfx_set_tv_flicker_filter(TV_FLICKER_FILTER_NORMAL);
				break;
		}
	}		
}

//---------------------------------------------------------------------------
// Checks whether or not tearing correction is necessary in the current
// settings, and sets a global variable accordingly.
// Tearing correction is needed in the following situation:
// 1. Video is vertically downscaled
// 2. Genlock to VIP is not active
//---------------------------------------------------------------------------

void
VidSetTearingCorrection(void)
{
	// Assumptions:
	//   1. A scaled video window is scaled vertically
	//   2. On scaled video window genlock to VIP signal is not done
	if (VidDisplay.current_window_size == VL_VIDEO_WINDOW_SCALED)
		VidDisplay.tearing_correction_active = 1;
	else
		VidDisplay.tearing_correction_active = 0;			
}


