/* NSC_LIC_BSD
 *
 * Revision 1.2
 *
 * National Semiconductor Corporation Open Source License for
 *
 *    TVLIB
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) July 19, 2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use by Licensee of this Software in source and
 * binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must either:
 *     (a) reproduce the above copyright notice, this list of conditions
 *         and the following disclaimer in the documentation and/or
 *         other materials provided with the distribution, or
 *     (b) if Licensee elects to redistribute any Software not in
 *         accordance with (a), then Licensee shall indemnify National
 *         Semiconductor Corporation and all copyright holders and
 *         contributors for any claims related to such Software (and pay
 *         all costs and damages, including all attorney fees)
 *
 *   * Neither the name of the National Semiconductor Corporation nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF
 * YOUR JURISDICTION. It is Licensee's responsibility to comply with
 * any export regulations applicable in Licensee's jurisdiction. Under
 * CURRENT (July 19, 2001) U.S. export regulations this software
 * is eligible for export from the U.S. and can be downloaded by or
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran,
 * Syria, Sudan, Afghanistan and any other country to which the U.S.
 * has embargoed goods and services.
 *
 * END_NSC_LIC_BSD */

/*
 * This file contains routines to use the I2C bus.
 *
 * I2C base address detection, initialization, and cleanup functions:
 * 
 *    i2c_get_acb_base_addresses
 *    tv_i2c_initialize
 *    tv_i2c_cleanup
 *
 * Linux system (kernel I2C support) I2C functions:
 *
 *    tv_i2c_attach_adapter
 *    tv_i2c_detach_client
 *    tv_i2c_detect_client
 *
 * I2C access functions:
 *
 *    tv_i2c_scan
 *    tv_i2c_cx241xx_read_byte
 *    tv_i2c_cx241xx_write_data
 *    tv_i2c_scart_write_data
 *    tv_i2c_saa7114_read_byte_data
 *    tv_i2c_saa7114_write_byte_data
 *    tv_i2c_tuner_read_byte
 *    tv_i2c_tuner_write_byte
 *    tv_i2c_tuner_write_byte_data
 *    tv_i2c_tuner_write
 *    tv_i2c_tda9851_read_byte
 *    tv_i2c_tda9851_write_byte
 *    tv_i2c_initialize
 *    tv_i2c_cleanup
 *
 * I2C utility functions:
 *
 *    acb_start_i2c_clock
 *    acb_stop_i2c_clock
 *    acb_write_bus_i2c 
 *    acb_write_i2c
 *    acb_start 
 *    acb_stop  
 *    acb_write_byte_i2c
 *    acb_reset_i2c
 *    acb_request_master
 *    acb_send_address
 *    acb_timeout          
 *    acb_start_timer     
 *    acb_abort_data
 *    acb_bus_recovery
 *    acb_stall_after_start
 *    acb_ack
 *    acb_read_byte_i2c 
 *    acb_read_i2c
 *    acb_read_bus_i2c
 *
 * Project: TVLIB
 * Author:  Peter Lohmann
 * Date:    April 10, 2001
 *---------------------------------------------------------------------*/

#ifdef LINUX
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif
#define __NO_VERSION__ 
#include <linux/module.h>
#include <linux/version.h>
#include <linux/time.h>
#endif
#include "tv_i2c.h"
#include "tv_os_geode.h"

#ifndef USE_SYSTEM_I2C
static ULONG	tv_i2c_start_time = 0;
#endif

void i2c_get_acb_base_addresses( WORD* acb1, WORD* acb2 )
{
	BYTE	h1 = 0;
	BYTE	l1 = 0;
	BYTE	h2 = 0;
	BYTE	l2 = 0;

	/* Select ACB1 -- LDN 5 */
	tv_os_io_write_byte( (WORD)0x015C, 0x07 );
	tv_os_io_write_byte( (WORD)0x015D, 0x05 );
	tv_os_io_write_byte( (WORD)0x015C, 0x60 );
	h1 = tv_os_io_read_byte( (WORD)0x015D );
	tv_os_io_write_byte( (WORD)0x015C, 0x61 );
	l1 = tv_os_io_read_byte( (WORD)0x015D );

	/* Select ACB2 -- LDN 6 */
	tv_os_io_write_byte( (WORD)0x015C, 0x07 );
	tv_os_io_write_byte( (WORD)0x015D, 0x06 );
	tv_os_io_write_byte( (WORD)0x015C, 0x60 );
	h2 = tv_os_io_read_byte( (WORD)0x015D );
	tv_os_io_write_byte( (WORD)0x015C, 0x61 );
	l2 = tv_os_io_read_byte( (WORD)0x015D );
	*acb1 = ( h1 << 8 ) | l1;
	*acb2 = ( h2 << 8 ) | l2;
}


// This is for the Linux I2C subsystem.
#ifdef USE_SYSTEM_I2C  

#define I2C_DRIVERID_GEODE	0x2222

#include <asm/io.h>
#include <linux/i2c.h>
#ifndef EXCLUDE_TUNER
#ifdef CYGNUS
#ifdef CYGNUS_PAL_TUNER
#define TUNER_ADD 0xC2
#define TUNER_STR "ALPS TMDZ2"
#else
#define TUNER_ADD 0xC6
#define TUNER_STR "ALPS TSBH1"
#endif
#else
#define TUNER_ADD 0xC6
#define TUNER_STR "Philips NTSC"
#endif
#endif
#define SAA7114_CHIPID  0x42
#define TDA9851_CHIPID  TDA9851_WRITE_ADR
#ifndef EXCLUDE_TUNER
static WORD normal_i2c[] = { SAA7114_CHIPID>>1, TUNER_ADD>>1,TDA9851_CHIPID>>1,I2C_CLIENT_END };
#else
static WORD normal_i2c[] = { SAA7114_CHIPID>>1,I2C_CLIENT_END };
#endif
static WORD normal_i2c_range[] = { I2C_CLIENT_END,I2C_CLIENT_END };
I2C_CLIENT_INSMOD;
int tv_i2c_attach_adapter(struct i2c_adapter *adapter);
int tv_i2c_detach_client(struct i2c_client *client);
int tv_i2c_detect_client(struct i2c_adapter *adapter, int address, WORD flags, int kind);

static struct i2c_driver geode_driver = 
{
	"Geode Video4Linux2 driver", /* name */
	I2C_DRIVERID_GEODE, /* id */
	I2C_DF_NOTIFY, /* flags */
	&tv_i2c_attach_adapter, /* attach_adapter */
	&tv_i2c_detach_client, /* detach_client */
	NULL, /* command */
	NULL, /* inc_use */
	NULL /* dec_use */
};

typedef struct tv_i2c_private_data 
{
	ULONG	data;
} tv_i2c_private_data_t;

struct i2c_client*	g_client_tuner = NULL;
struct i2c_client*	g_client_decoder = NULL;
struct i2c_client*	g_client_tda9851 = NULL;

int tv_i2c_initialize()
{
	int	res;
	if( ( res = i2c_add_driver( &geode_driver ) ) ) 
	{
		tv_os_err_msg("tv_i2c_initialize(): i2c_add_driver() failed.\n");
		tv_i2c_cleanup();
	}
	tv_os_dbg_msg("tv_i2c_initialize(): i2c_add_driver() succeeded.\n");
	return res;
}

int tv_i2c_cleanup()
{
	int	res;
	if( g_client_tuner )
	{
		i2c_detach_client( g_client_tuner );
		g_client_tuner = NULL;
	}
	if( g_client_decoder )
	{
		i2c_detach_client( g_client_decoder );
		g_client_decoder = NULL;
	}
	if( g_client_tda9851 )
	{
		i2c_detach_client( g_client_tda9851 );
		g_client_tda9851 = NULL;
	}
	if( ( res = i2c_del_driver( &geode_driver ) ) ) 
	{
		tv_os_err_msg("tv_i2c_cleanup(): i2c_del_driver()  failed.n");
	}
	return res;
}

int tv_i2c_attach_adapter(struct i2c_adapter *adapter)
{
	tv_os_dbg_msg("tv_i2c_attach_adapter\n");
	return i2c_probe( adapter, &addr_data, &tv_i2c_detect_client );
}

int tv_i2c_detach_client(struct i2c_client *client)
{
	int err;

	if ( ( err = i2c_detach_client( client ) ) ) 
	{
		tv_os_err_msg("tv_i2c_detach_client(): i2c_detach_client() failed.\n");
		return err;
	}

	tv_os_kfree( client );
	return 0;
}

int tv_i2c_detect_client(struct i2c_adapter *adapter, int address, WORD flags, int kind)
{
	static int 	geode_id = 0;
	int 		err = 0;
	struct i2c_client*	new_client = NULL;
	struct tv_i2c_private_data*	data = NULL;
	
	/* allocate space for i2c_client structure. */
	new_client = tv_os_kmalloc( sizeof(struct i2c_client) + sizeof(struct tv_i2c_private_data) );
	if( new_client == NULL )
	{
		err = -ENOMEM;
		return err;
	}

	if( strcmp(adapter->name,"NSC SC1200 ACB adapter (ACB1)") == 0 )
	{
#ifndef EXCLUDE_TUNER
		if( ( address << 1 ) == TUNER_ADD )
		{
			printk("Found "TUNER_STR" tuner\n");
			g_client_tuner = new_client;
		}
		else if( ( address << 1 ) == TDA9851_CHIPID )
		{
			printk("Found TDA 9851 FM de-mux\n");
			g_client_tda9851 = new_client;
		}
#endif		
	}
	else if( strcmp(adapter->name,"NSC SC1200 ACB adapter (ACB2)") == 0 )
	{
		if( ( address << 1 ) == SAA7114_CHIPID )
		{
			printk("Found SAA714 decoder\n");
			g_client_decoder = new_client;
		}
	}
	else
	{
		tv_os_kfree(new_client);
		tv_os_dbg_msg("No Client detected\n");
		return 0;
	}

	/* This is tricky, but it will set the data to the right value. */
	new_client->data = new_client + 1;
	data = (struct tv_i2c_private_data *) (new_client + 1);

	new_client->addr = address;
	new_client->data = data;
	new_client->adapter = adapter;
	new_client->driver = &geode_driver;
	new_client->flags = 0;
	strcpy(new_client->name, "geode client");

	new_client->id = geode_id++; /* Automatically unique */

	/* Tell the i2c layer a new client has arrived */
	if( ( err = i2c_attach_client( new_client ) ) ) 
	{
		tv_os_dbg_msg("tv_i2c_detect_client(): i2c_attach_client() failed\n");
		i2c_detach_client(new_client);
		tv_os_kfree(new_client);
		return err;
	}
	tv_os_dbg_msg("tv_i2c_detect_client(): i2c_attach_client() succeeded\n");

	return 0;
}

BYTE tv_i2c_saa7114_read_byte_data( BYTE reg )
{
	if(g_client_decoder)
	return i2c_smbus_read_byte_data( g_client_decoder, reg );
	else
		return 0xFF;
}

BYTE tv_i2c_saa7114_write_byte_data( BYTE reg, BYTE value )
{
	if(g_client_decoder)
	return i2c_smbus_write_byte_data( g_client_decoder, reg, value );
	else
		return 0xFF;
}

// We have to do it this way because the bus writes do not seem to work.
BYTE tv_i2c_tuner_write_byte_data( BYTE reg, WORD value )
{
	if(g_client_tuner)
	return i2c_smbus_write_byte_data( g_client_tuner, reg, value );
	else
		return 0xFF;
}

BYTE tv_i2c_tuner_write( BYTE addr, WORD divider, WORD control_data )
{
	tv_i2c_tuner_write_byte_data((BYTE)((divider >> 8)&0x7f), divider&0xff );
	tv_i2c_tuner_write_byte_data((BYTE)((control_data>>8)&0xff),control_data&0xFF);
	return 1;
}

BYTE tv_i2c_tuner_read_byte(BYTE addr)
{
	if(g_client_tuner)
	return i2c_smbus_read_byte( g_client_tuner );
	else
		return 0xFF;
}

BYTE tv_i2c_tuner_write_byte( WORD value )
{
	if(g_client_tuner)
	return i2c_smbus_write_byte( g_client_tuner, value );
	else
		return 0xFF;
}

BYTE tv_i2c_tda9851_read_byte()
{
	if( g_client_tda9851)
	return i2c_smbus_read_byte( g_client_tda9851 );
	else
		return 0xff;
}

BYTE tv_i2c_tda9851_write_byte( BYTE value )
{
	if(g_client_tda9851)
	return i2c_smbus_write_byte( g_client_tda9851, value );
	else
		return 0xff;
}

#else	// !USE_SYSTEM_I2C

// Give them some reasonable defaults.
WORD	geode_acb1_base_addr = GEODE_ACB_BASE_ADDR;
WORD	geode_acb2_base_addr = GEODE_ACB2_BASE_ADDR;

unsigned long	tv_i2c_timeout_cnt = 0;

int tv_i2c_initialize()
{
	tv_i2c_timeout_cnt = 0;
	i2c_get_acb_base_addresses( &geode_acb1_base_addr, &geode_acb2_base_addr );
	return 0;
}

int tv_i2c_cleanup()
{
	return 0;
}

void tv_i2c_scan()
{
	unsigned char i = 0;
	tv_os_dbg_msg( "Scanning I2C bus...\n" );
	for( i = 0; i < 0x7F; i++ )
	{
		if( acb_request_master( geode_acb1_base_addr ) == 0 ) 
		{
			continue;
		}
		acb_ack( geode_acb1_base_addr, 1, 0 );
		acb_send_address( geode_acb1_base_addr, i << 1 );
		if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
		{
			// No device present at this address?
			acb_bus_recovery( geode_acb1_base_addr );
		}
		else
		{
			tv_os_dbg_msg( "bus 1: device at 0x%02X\n", i << 1 );
		}
		acb_stop( geode_acb1_base_addr );
	}
	for( i = 0; i < 0x7F; i++ )
	{
		if( acb_request_master( geode_acb2_base_addr ) == 0 ) 
		{
			continue;
		}
		acb_ack( geode_acb2_base_addr, 1, 0 );
		acb_send_address( geode_acb2_base_addr, i << 1 );
		if( !acb_ack( geode_acb2_base_addr, 0, 0 ) )
		{
			// No device present at this address?
			acb_bus_recovery( geode_acb2_base_addr );
		}
		else
		{
			tv_os_dbg_msg( "bus 2: device at 0x%02X\n", i << 1 );
		}
		acb_stop( geode_acb2_base_addr );
	}
}

int tv_i2c_cx241xx_read_data( BYTE ucReg, BYTE* pucData )
{
	if( acb_read_i2c_verify( geode_acb1_base_addr, CX241XX_READ_ADR, ucReg, pucData ) )
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

int tv_i2c_cx241xx_write_data( BYTE ucReg, BYTE ucVal )
{							 
	return acb_write_i2c_verify( geode_acb1_base_addr, CX241XX_WRITE_ADR, ucReg, ucVal );		    
}

BYTE tv_i2c_scart_write_data( BYTE ucVal1, BYTE ucVal2 )
{
	return acb_write_i2c( geode_acb2_base_addr, SCART_WRITE_ADR, ucVal1, ucVal2 );		    
}

BYTE tv_i2c_tda9851_read_byte()
{
	return acb_read_bus_i2c( geode_acb1_base_addr, TDA9851_READ_ADR );
}

BYTE tv_i2c_tda9851_write_byte( BYTE value )
{							 
	if( acb_request_master( geode_acb1_base_addr ) == 0 ) 
	{
		tv_os_err_msg( "acb_request_master() failed\n" );
		return 0;
	}
	acb_ack( geode_acb1_base_addr, 1, 0 );
	acb_send_address( geode_acb1_base_addr, TDA9851_WRITE_ADR );
	if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
	{
		return 0;
	}
	acb_write_byte_i2c( geode_acb1_base_addr, value );
	if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
	{
		return 0;
	}
	acb_stop( geode_acb1_base_addr );
	return 1;
}

BYTE tv_i2c_saa7114_read_byte_data( BYTE reg )
{
	return acb_read_i2c( geode_acb2_base_addr, DEC_READ_ADR, reg );		    
}

BYTE tv_i2c_saa7114_write_byte_data( BYTE reg, BYTE value )
{
	return acb_write_i2c( geode_acb2_base_addr, DEC_WRITE_ADR, reg, value );		    
}

BYTE tv_i2c_tuner_read_byte( BYTE tuner_i2c_read_addr )
{
	return acb_read_bus_i2c( geode_acb1_base_addr, tuner_i2c_read_addr );
}

BYTE tv_i2c_tuner_write( BYTE tuner_i2c_write_addr, WORD divider, WORD control_data )
{
	if( acb_request_master( geode_acb1_base_addr ) == 0 ) 
	{
		tv_os_err_msg( "acb_request_master() failed\n" );
		return 0;
	}
	acb_ack( geode_acb1_base_addr, 1, 0 );
	acb_send_address( geode_acb1_base_addr, tuner_i2c_write_addr );
	if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
	{
		return 0;
	}
	acb_write_byte_i2c( geode_acb1_base_addr, (BYTE)( divider >> 8 ) );
	if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
	{
		return 0;
	}
	acb_write_byte_i2c( geode_acb1_base_addr, (BYTE)divider );
	if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
	{
		return 0;
	}
	acb_write_byte_i2c( geode_acb1_base_addr, (BYTE)( control_data >> 8 ) );
	if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
	{
		return 0;
	}
	acb_write_byte_i2c( geode_acb1_base_addr, (BYTE)control_data );
	if( !acb_ack( geode_acb1_base_addr, 0, 0 ) )
	{
		return 0;
	}
	acb_stop( geode_acb1_base_addr );
	return 1;
}


int acb_write_bus_i2c( WORD bus, BYTE chipadr, BYTE data )
{
	if( acb_request_master( bus ) == 0 ) 
	{
		return 0;
	}
	
	acb_ack( bus, 1, 0 );
	acb_send_address( bus, (BYTE)( chipadr & 0xFE ) );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	acb_write_byte_i2c( bus, data );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	acb_stop( bus );
	
	return 1;
}


int acb_write_i2c_verify( WORD bus, BYTE chipadr, BYTE subadr, BYTE data )
{
	if( acb_request_master( bus ) == 0 ) 
	{
		return 0;
	}
	
	acb_ack( bus, 1, 0 );

	if( !acb_send_address_verify( bus, (BYTE)( chipadr & 0xFE ) ) )
	{
		return 0;
	}
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	if( !acb_write_byte_i2c_verify( bus, subadr ) )
	{
		return 0;
	}
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	if( !acb_write_byte_i2c_verify( bus, data ) )
	{
		return 0;
	}
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	acb_stop( bus );
	return 1;
}

int acb_write_i2c( WORD bus, BYTE chipadr, BYTE subadr, BYTE data )
{
	if( acb_request_master( bus ) == 0 ) 
	{
		return 0;
	}
	
	acb_ack( bus, 1, 0 );
	acb_send_address( bus, (BYTE)( chipadr & 0xFE ) );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	acb_write_byte_i2c( bus, subadr );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	acb_write_byte_i2c( bus, data );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	acb_stop( bus );
	return 1;
}

int acb_timeout()
{
	ULONG curr_time = tv_os_get_tick_count();
	
	if( ( curr_time - tv_i2c_start_time ) >= 10000 ) 
	{
		tv_i2c_timeout_cnt++;
		acb_reset_i2c( geode_acb1_base_addr );
		acb_reset_i2c( geode_acb2_base_addr );
		tv_os_err_msg("****** acb_timeout() ******\n");
		return 1;
	}
	else
	{
		return 0;
	}
}

void acb_start_timer()
{
	tv_i2c_start_time = tv_os_get_tick_count();
}

void acb_start_i2c_clock( WORD bus )
{
	BYTE reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL2 ) );
	reg = reg | 0x01;
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL2 ), reg );
}

void acb_stop_i2c_clock( WORD bus )
{
	BYTE reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL2 ) );
	reg = reg & 0xFE;
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL2 ), reg );
}

void acb_start( WORD bus )
{
	BYTE reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ) );
	reg = reg | 0x01;
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ), reg );
}

void acb_stop( WORD bus )
{
	BYTE reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ) );
	reg = reg | 0x02;
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ), reg );
}

void acb_abort_data( WORD bus )
{
	BYTE reg;
	acb_stop( bus );
	reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ) );
	reg = reg | 0x10;
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ), reg );
}

void acb_bus_recovery( WORD bus )
{
	acb_abort_data( bus );
	acb_reset_i2c( bus );
}

void acb_stall_after_start( WORD bus, int state )
{
	BYTE reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ) );
	if( state == 1 )
	{
		reg = reg | 0x80;
	}
	else
	{
		reg = reg & 0x7f;
	}
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ), reg );
	
	if( state == 0 )
	{
		reg = READ_PORT_UCHAR( (BYTE*)( bus+ GEODE_ACBST ) );
		reg = reg | 0x08;
		WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ), reg );
	}
}

int acb_send_address_verify( WORD bus, BYTE cData )
{
	BYTE reg;
	
	acb_stall_after_start( bus, 1 );
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBSDA ), cData );
	acb_start_timer();

	while( 1 ) 
	{
		reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ) );
		if( ( reg & 0x38 ) != 0 )
		{
			break;
		}
		if( acb_timeout() )
		{
			acb_bus_recovery( bus );
			return 0;
		}
	}
	if( ( reg & 0x20 ) != 0 ) 
	{		
		acb_bus_recovery( bus );
		return 0;
	}
	if( ( reg & 0x10 ) != 0 ) 
	{		
		acb_abort_data( bus );
		return 0;
	}
	acb_stall_after_start( bus, 0 );
	return 1;
}

void acb_send_address( WORD bus, BYTE cData )
{
	BYTE reg;
	
	acb_stall_after_start( bus, 1 );
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBSDA ), cData );
	acb_start_timer();

	while( 1 ) 
	{
		reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ) );
		if( ( reg & 0x38 ) != 0 )
		{
			break;
		}
		if( acb_timeout() )
		{
			acb_bus_recovery( bus );
			return;
		}
	}
	if( ( reg & 0x20 ) != 0 ) 
	{		
		acb_bus_recovery( bus );
		return;
	}
	if( ( reg & 0x10 ) != 0 ) 
	{		
		acb_abort_data( bus );
		return;
	}
	acb_stall_after_start( bus, 0 );
	return;
}

int acb_ack( WORD bus, int fPut, int negAck )
{
	BYTE reg;
	
	if( fPut ) 
	{		
		if( !negAck ) 
		{
			reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ) );
			reg = reg & 0xE7;
			WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ), reg );
		}
		else 
		{
			reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ) );
			reg = reg | 0x10;
			WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ), reg );
		}
	} 
	else
	{		
		acb_start_timer();
		while( 1 ) 
		{
			reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ) );
			if( ( reg & 0x70 ) != 0 )
			{
				break;
			}
			if( acb_timeout() )
			{
				acb_bus_recovery( bus );
				return 0;
			}
		}
		if( ( reg & 0x20 ) != 0 ) 
		{
			acb_bus_recovery( bus );
			return 0;
		}
		if( ( reg & 0x10 ) != 0 )
		{		
			acb_abort_data( bus );
			return 0;
		}
	}
	return 1;
}                

BYTE acb_read_byte_i2c( WORD bus )
{
	BYTE		cData;
	BYTE	reg;
	
	acb_start_timer();
	while( 1 ) 
	{
		reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ) );
		if( ( reg & 0x60 ) != 0 )
		{
			break;
		}
		if( acb_timeout() )
		{
			acb_bus_recovery( bus );
			return (BYTE)0xff;
		}
	}
	if( ( reg & 0x20 ) != 0 ) 
	{
		acb_bus_recovery( bus );
		return (BYTE)0xff;
	}
	
	acb_stop_i2c_clock( bus );
	cData = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBSDA ) );
	acb_start_i2c_clock( bus );
	
	return cData;
}

int acb_read_i2c_verify( WORD bus, BYTE chipadr, BYTE subadr, BYTE* pucData )
{
	BYTE data;
	
	if( acb_request_master( bus ) == 0 ) 
	{
		return 0;
	}
	
	acb_ack( bus, 1, 0 );

	if( !acb_send_address_verify( bus, (BYTE)( chipadr & 0xFE ) ) )
	{
		return 0;
	}
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	if( !acb_write_byte_i2c_verify( bus, subadr ) )
	{
		return 0;
	}
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	acb_start( bus );
	
	acb_ack( bus, 1, 1 );
	if( !acb_send_address_verify( bus, (BYTE)( chipadr | 0x01 ) ) )
	{
		return 0;
	}
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return 0;
	}
	
	acb_ack( bus, 1, 1 );
	data = acb_read_byte_i2c( bus );
	
	acb_stop( bus );

	*pucData = data;

	return 1;
}

BYTE acb_read_i2c( WORD bus, BYTE chipadr, BYTE subadr )
{
	BYTE data;
	
	if( acb_request_master( bus ) == 0 ) 
	{
		return (BYTE)0xff;
	}
	
	acb_ack( bus, 1, 0 );

	acb_send_address( bus, (BYTE)( chipadr & 0xFE ) );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return (BYTE)0xff;
	}
	
	acb_write_byte_i2c( bus, subadr );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return (BYTE)0xff;
	}
	
	acb_start( bus );
	
	acb_ack( bus, 1, 1 );
	acb_send_address( bus, (BYTE)( chipadr | 0x01 ) );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return (BYTE)0xff;
	}
	
	acb_ack( bus, 1, 1 );
	data = acb_read_byte_i2c( bus );
	
	acb_stop( bus );

	return data;
}

BYTE acb_read_bus_i2c( WORD bus, BYTE chipadr )
{
	BYTE data;
	
	if( acb_request_master( bus ) == 0 ) 
	{
		return (BYTE)0xff;
	}
	
	acb_ack( bus, 1, 0 );
	acb_send_address( bus, (BYTE)( chipadr & 0xFE ) );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return (BYTE)0xff;
	}
	
	acb_start( bus );
	
	acb_ack( bus, 1, 1 );
	acb_send_address( bus, (BYTE)( chipadr | 0x01 ) );
	
	if( !acb_ack( bus, 0, 0 ) )
	{
		return (BYTE)0xff;
	}
	
	acb_ack( bus, 1, 1 );
	data = acb_read_byte_i2c( bus );
	
	acb_stop( bus );
	return data;
}

int acb_write_byte_i2c_verify( WORD bus, BYTE cData )
{
	BYTE reg;
	
	acb_start_timer();
	while( 1 ) 
	{
		reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ) );
		if( ( reg & 0x70 ) != 0 )	
		{
			break;
		}
		if( acb_timeout() )
		{
			acb_bus_recovery( bus );
			return 0;
		}
	}
	if( ( reg & 0x20 ) != 0 ) 
	{		
		acb_bus_recovery( bus );
		return 0;
	}
	if( ( reg & 0x10 ) != 0 ) 
	{		
		acb_abort_data( bus );
		return 0;
	}
	
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBSDA ), cData );
	return 1;
}

void acb_write_byte_i2c( WORD bus, BYTE cData )
{
	BYTE reg;
	
	acb_start_timer();
	while( 1 ) 
	{
		reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ) );
		if( ( reg & 0x70 ) != 0 )	
		{
			break;
		}
		if( acb_timeout() )
		{
			acb_bus_recovery( bus );
			return;
		}
	}
	if( ( reg & 0x20 ) != 0 ) 
	{		
		acb_bus_recovery( bus );
		return;
	}
	if( ( reg & 0x10 ) != 0 ) 
	{		
		acb_abort_data( bus );
		return;
	}
	
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBSDA ), cData );
}

void acb_reset_i2c( WORD bus )
{
	BYTE reg;
	
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL2 ), 0x70 );
	
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL1 ), 0x0 );
	
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBADDR ), 0x0 );
	
	reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL2 ) );
	reg = reg | 0x01;
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCTL2 ), reg );
	
	acb_stop( bus );
	
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ), 0x38 );
	
	reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCST ) );
	reg = reg | 0x02;
	WRITE_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBCST ), reg );
}

int acb_request_master( WORD bus )
{
	BYTE reg;
	
	acb_start( bus );
	acb_start_timer();
	while( 1 ) 
	{
		reg = READ_PORT_UCHAR( (BYTE*)( bus + GEODE_ACBST ) );
		if( ( reg & 0x60 ) != 0 )
		{
			break;
		}
		if( acb_timeout() )
		{
			acb_bus_recovery( bus );
			return 0;
		}
	}
	if( ( reg & 0x20 ) != 0 ) 
	{		
		acb_abort_data( bus );
		return 0;
	}
	if( ( reg & 0x02 ) == 0 ) 
	{		
		acb_abort_data( bus );
		return 0;
	}
	return 1;
}

#endif

