/*
 *
 * National Semiconductor Corporation Open Source License for 
 *
 *    NSC Analog Video Middleware
 *
 * (BSD License with Export Notice)
 *
 * Copyright (c) 2001
 * National Semiconductor Corporation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution. 
 *
 *   * Neither the name of the National Semiconductor Corporation nor 
 *     the names of its contributors may be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * NATIONAL SEMICONDUCTOR CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER 
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE,
 * INTELLECTUAL PROPERTY INFRINGEMENT, OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * EXPORT LAWS: THIS LICENSE ADDS NO RESTRICTIONS TO THE EXPORT LAWS OF 
 * YOUR JURISDICTION. It is licensee's responsibility to comply with 
 * any export regulations applicable in licensee's jurisdiction. Under 
 * CURRENT (2001) U.S. export regulations this software 
 * is eligible for export from the U.S. and can be downloaded by or 
 * otherwise exported or reexported worldwide EXCEPT to U.S. embargoed 
 * destinations which include Cuba, Iraq, Libya, North Korea, Iran, 
 * Syria, Sudan, Afghanistan and any other country to which the U.S. 
 * has embargoed goods and services. 
 *
 */

#include <stdlib.h>
#include "anlg_video.h"

/* Our signal handler for the VBI signal */
void signal_handler(int signum);

//#define DEBUG
static VMW_VIDEO * pvideo = NULL; /* pointer to global device context */

/*
 *	Algorithm:
 *		First we check if we have a valid device name. We then open
 * 		the device either in the normal mode or the NOIO mode
 *		depending on the value of the flag. If the open is
 *		successful, the resulting file descriptor is stored in the
 *		device context and the status is set to VIDEO_OPENED.
 *	
 */
u32 VMW_AVS_VideoInit()
{ 
	if(pvideo)
	{
		MSG("pvideo not NULL (already opened?)");
		return VMW_AVSE_INVSTAT ;
	}
	/* initialize the device context */
	pvideo =  (VMW_VIDEO *)malloc(sizeof(VMW_VIDEO));	
	if(pvideo == NULL)
	{
		MSG("Unable to allocate memory for device context\n");
		return VMW_AVSE_INIT;
	}

	memset(pvideo,0,sizeof(VMW_VIDEO));
	
	/* initialize the state */
	pvideo->callback = NULL;
	pvideo->callback_data = NULL;
	pvideo->vbi_read_buffer_length=0;

	/* lets open all the necessary device files now */
	/* first the video device */
	DBG("Opening VIDEO\n");
	pvideo->fd = open(DEV_FILE_VIDEO,O_RDWR);
	
	/* If we fail here, we might as well give up */
	if(pvideo->fd < 0)
	{
		MSG("Error opening " DEV_FILE_VIDEO"\n");
		return VMW_AVSE_OPEN;
	}
	
#ifndef NO_NOIO
	DBG("Opening VIDEO NOIO\n");
	/* then the noio video */
	pvideo->fd_noio = open(DEV_FILE_VIDEO, O_RDWR|O_NOIO);
#endif	

#ifndef NO_VBI
	DBG("Opening VBI \n");
	/* then we open the vbi device */
	pvideo->fd_vbi = open(DEV_FILE_VBI, O_RDWR);
#ifndef NO_NOIO	
	DBG("Opening VBI NOIO \n");
	pvideo->fd_vbi_noio = open(DEV_FILE_VBI, O_RDWR|O_NOIO);
#endif /* NO_NOIO */
#endif /* NO_VBI  */

	/* then the vout device */
	DBG("Opening VOUT \n");
	pvideo->fd_vout = open(DEV_FILE_VOUT,O_RDWR);
#ifndef NO_NOIO	
	DBG("Opening VOUT NOIO \n");
	pvideo->fd_vout_noio = open(DEV_FILE_VOUT,O_RDWR|O_NOIO);
#endif /* NO_NOIO */

	/* lets now check and set the status bits to indicate all the
	 successful opens. */

	if(pvideo->fd >= 0)
	{
		pvideo->status |= VIDEO_OPENED;
		DBG("Video opened\n");
	}
	if(pvideo->fd_noio >= 0)
	{
		pvideo->status |= VIDEO_OPENED_NOIO;	
		DBG("Video opened NOIO\n");
	}
	if(pvideo->fd_vbi >= 0)
	{
		pvideo->status |= VBI_OPENED;	
		DBG("VBI opened\n");
	}
	if(pvideo->fd_vbi_noio >= 0)
	{
		pvideo->status |= VBI_OPENED_NOIO;	
		DBG("VBI opened NOIO\n");
	}
	if(pvideo->fd_vout >= 0)
	{
		pvideo->status |= VOUT_OPENED;	
		DBG("vout opened \n");
	}
	if(pvideo->fd_vout_noio >= 0)
	{
		pvideo->status |= VOUT_OPENED_NOIO;	
		DBG("vout opened NOIO\n");
	}

	if(pvideo->status == 0)
	{
		/*this means, we couldn't open anything */
		return VMW_AVSE_OPEN;
	}
	
	return 0 ;  
}


/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_QUERYCAP ioctl.
 *	
 */
u32 VMW_AVS_GetVideoCapabilities(VIDEO_CAPABILITY *pvidcap )
{
	/* check if we're opened */
	if( (pvideo != NULL) && (pvideo->status & VIDEO_OPENED)  ) 
	{
		DBG("VMW_AVS_GetVideoCapabilities\n");
		return DO_IOCTL(pvideo->fd, VIDIOC_QUERYCAP, pvidcap );
	}
	else 
	{
		DBG("Invalid Status in VMW_GetVideoCapabilities\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	deprecated
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call either the
 *		VIDIOC_ENUM_PIXFMT ioctl or VIDEO_ENUM_FBUFFMT ioctl
 *		depending on the parameter passed.
 *	
 */
u32 VMW_AVS_EnumFormat(int type, VIDEO_FMTDESC * pvidfmtdesc )
{
	/* check if we're opened */
	if( (pvideo == NULL) || (!(pvideo->status & VIDEO_OPENED )) ) 
	{
		MSG("Device not opened\n");
		return VMW_AVSE_INVSTAT;
	}

	switch(type)
	{
		case ENUM_PIXEL_FMT:
			return ( DO_IOCTL(pvideo->fd,VIDIOC_ENUM_PIXFMT,
				       pvidfmtdesc ) );	
			break; 

		case ENUM_FRBUFF_FMT:
			return ( DO_IOCTL (pvideo->fd,VIDIOC_ENUM_FBUFFMT,
			                pvidfmtdesc ) );      
			break;
		
		default:
			MSG("Unknown type\n");
			break;
				
	}
	return VMW_AVSE_INVSTAT;  

}

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_ENUM_PIXFMT ioctl.
 *	
 */

#if 0
u32 VMW_AVS_EnumPixelFormat(VIDEO_FMTDESC *pvidfmtdesc )
{
	/* check if we're opened */
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVS_EnumPixelFormat\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_ENUM_PIXFMT,
				pvidfmtdesc );
	}
	else 
	{
		DBG("Invalid Status in VMW_AVS_EnumPixelFormat\n");
		return VMW_AVSE_INVSTAT;
	}
}

#endif

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_ENUM_FBUFFMT ioctl.
 */
u32 VMW_AVS_EnumFrameBufferFormat(VIDEO_FMTDESC *pvidfmtdesc )
{
	/* check if we're opened */
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_EnumFrameBufferFormat\n");
		return DO_IOCTL(pvideo->fd, VIDIOC_ENUM_FBUFFMT,
				pvidfmtdesc );
	}
	else 
	{
		DBG("Invalid Status in VMW_AVS_EnumFrameBufferFormat\n");
		return VMW_AVSE_INVSTAT;
	}

}


/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_G_FMT ioctl.
 */
u32 VMW_AVC_GetCaptureFormat(VIDEO_FORMAT *pformat)
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVC_GetCaptureFormat\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_G_FMT,
				pformat );
	}
	else 
	{
		DBG("Invalid Status in VMW_AVC_GetCaptureFormat\n");
		return VMW_AVSE_INVSTAT;
	}
}


/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_S_FMT ioctl.
 */
u32 VMW_AVC_SetCaptureFormat(VIDEO_FORMAT * pformat )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVC_SetCaptureFormat\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_S_FMT,
				pformat );
	}
	else 
	{
		DBG("Invalid Status VMW_SetCaptureFormat\n");
		return VMW_AVSE_INVSTAT;
	}
	
}


u32 VMW_AVS_Read(char * buffer, size_t count ) 
{
	size_t len;
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_Read\n");
		len = read(pvideo->fd, buffer, count );
		if ( len  < count ) // incomplete read...failed!
		{
			return VMW_AVSE_READ ;
		}
		else return 0;
	}
	else
	{
		DBG("Invalid Status in VMW_AVS_Read\n");
		return VMW_AVSE_INVSTAT;
	}
}


/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_ENUMINPUT ioctl.
 */
u32 VMW_AVS_EnumVideoInput( VIDEO_INPUT *pvidinp )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVS_EnumVideoInput\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_ENUMINPUT,
				pvidinp );
	}
	else 
	{
		DBG("Invalid Status VMW_AVS_EnumVideoInput\n");
		return VMW_AVSE_INVSTAT;
	}
}


/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_G_INPUT ioctl.
 */
u32 VMW_AVS_GetCurrentVideoInput(int * pindex )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVS_GetCurrentVideoInput\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_G_INPUT,
				pindex );
	}
	else 
	{
		DBG("Invalid Status in VMS_AVS_GetCurrentVideoInput\n");
		return VMW_AVSE_INVSTAT;
	}
}


/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call either the
 *		VIDIOC_STREAMON or the VIDIOC_STREAMOFF ioctl depend
 *		on the parameter flag.
 */
u32 VMW_AVS_EnableCapture(int enable)
{
        int dummy = 1;
	DBG("in enable captuer\n");
	if( (pvideo != NULL) && (pvideo->status & VIDEO_OPENED)  ) 
	{
	if (enable){
		return( DO_IOCTL(pvideo->fd, VIDIOC_STREAMON,&dummy));
	}
	else{
		return( DO_IOCTL(pvideo->fd, VIDIOC_STREAMOFF,&dummy));
	}
	}
	else
	{
                MSG("Video device not opened\n");
                return  VMW_AVSE_INVSTAT;
        }
}

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_S_INPUT ioctl.
 */
u32 VMW_AVS_SetCurrentVideoInput(int * pindex )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVS_SetCurrentVideoInput\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_S_INPUT,
				pindex );
	}
	else 
	{
		DBG("Invalid Status in VMW_AVS_SetCurrentVideoInput\n");
		return VMW_AVSE_INVSTAT;
	}

}

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_G_TUNER ioctl.
 *	
 */
u32 VMW_AVT_GetTunerCapabilities( VIDEO_TUNER *ptuner )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) )
	{
		DBG("VMW_AVT_GetTunerCapabilities\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_G_TUNER,
				ptuner);
	}
	else 
	{
		DBG("Invalid Status in VMW_AVT_GetTunerCapabilities\n");
		return VMW_AVSE_INVSTAT;
	}

}

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_S_TUNER ioctl.
 */
u32 VMW_AVT_SetTunerCapabilities( VIDEO_TUNER *ptuner )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_SetTunerCapabilities\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_S_TUNER,
				ptuner);
	}
	else 
	{
		DBG("Invalid Status VMW_AVT_SetTunerCapabilities\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	Algorithm:
 *	 	The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_G_FREQ ioctl.
 */
u32 VMW_AVT_GetTunerFrequency(unsigned long * pfrequency )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVT_GetTunerFrequency\n");
		return DO_IOCTL(pvideo->fd_noio,VIDIOC_G_FREQ,
				pfrequency);
	}
	else 
	{
		DBG("VMW_AVT_GetTunerFrequency\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_S_FREQ ioctl.
 */
u32 VMW_AVT_SetTunerFrequency(unsigned long * pfrequency )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVT_SetTunerFrequency\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_S_FREQ,
				pfrequency);
	}
	else 
	{
		DBG("Invalid Status in VMW_AVT_SetTunerFrequency\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *	 	VIDIOC_G_AUDIO ioctl.
 *	
 */
u32 VMW_AVA_GetAudioCapabilities(VIDEO_AUDIO *paudio_cap )
{
	/*check if we're opened */
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVA_GetAudioCapabilities\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_G_AUDIO,
				paudio_cap);
	}
	else 
	{
		DBG("VMW_AVA_GetAudioCapabilities\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	Algorithm:
 *		The status of the video context is checked for the
 *		VIDEO_OPENED flag. If this is set, we call the
 *		VIDIOC_S_AUDIO ioctl.
 */

u32 VMW_AVA_SetAudioCapabilities(VIDEO_AUDIO *paudio_cap )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVA_SetAudioCapabilities\n");
		return DO_IOCTL(pvideo->fd,VIDIOC_S_AUDIO,
				paudio_cap);
	}
	else 
	{
		DBG("VMW_AVA_SetAudioCapabilities\n");
		return VMW_AVSE_INVSTAT;
	}
}


/*
 *	Algorithm:
 *		Fill up a video_window structure and  call the VIDIOC_S_WIN
 *		ioctl.
 *	Side Effect: this function also sets the capture format  
 */
u32 VMW_AVS_SetVideoWindow(int x, int y, int width, int height )
{
	VIDEO_WINDOW vid_win;
	VIDEO_FORMAT fmt;

	vid_win.x = x ;
	vid_win.y = y;
	vid_win.width = width;
	vid_win.height = height;
	fmt.fmt.pix.width = width;
	fmt.fmt.pix.height = height;

	
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVS_SetVideoWindow\n");
		VMW_AVC_SetCaptureFormat(&fmt);
		VMW_AVS_EnableCapture(1);
		return DO_IOCTL(pvideo->fd_vout,VIDIOC_S_WIN,
				&vid_win);
	}
	else 
	{
		DBG("Invalid Status in VMW_AVS_SetVideoWindow\n");
		return VMW_AVSE_INVSTAT;
	}
		
}


/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_QUERYCAP ioctl.
 */
u32 VMW_AVV_GetVBICapabilities(VIDEO_CAPABILITY * pvidcap)
{
	if((pvideo != NULL) && (pvideo->status & VBI_OPENED) ) 
	{
		DBG("VMW_AVV_GetVBICapabilities\n");
		return DO_IOCTL(pvideo->fd_vbi, VIDIOC_QUERYCAP, pvidcap );
	}
	else 
	{
		MSG("VME_GetVBICapabilities: vbi device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
	
}


/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_G_FMT ioctl.
 */
u32 VMW_AVV_GetVBIFormat(VIDEO_FORMAT *pformat )
{
	if((pvideo != NULL) && (pvideo->status & VBI_OPENED) )
	{
		DBG("VMW_AVV_GetVBIFormat\n");
		return DO_IOCTL(pvideo->fd_vbi, VIDIOC_G_FMT, pformat );
	}
	else 
	{
		MSG("VME_AVV_GetVBICapabilities: vbi device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 * Description:
 *	enable or disable DVIP depending on the input
 */
u32 VMW_AVS_EnableDVIP(int enable)
{
	MSG("enabling DVIP\n");
	return DO_IOCTL(pvideo->fd_vout, GEODE_IOCTL_S_DVIP,&enable);
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_S_FMT ioctl.
 */
u32 VMW_AVV_SetVBIFormat(VIDEO_FORMAT *pformat )
{
	if((pvideo != NULL) && (pvideo->status & VBI_OPENED) ) 
	{
		DBG("VMW_AVV_GetVBIFormat\n");
		
		if(!ioctl(pvideo->fd_vbi, VIDIOC_S_FMT, pformat ))
		{	
			/* set the VBI read length */
			pvideo->vbi_read_buffer_length = 
					pformat->fmt.vbi.samples_per_line *
					 (pformat->fmt.vbi.count[0] +
					 pformat->fmt.vbi.count[1]) ;
		}
		else return VMW_AVSE_IOCTL;			 
	}
	else 
	{
		MSG("VME_AVV_GetVBICapabilities: vbi device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_ENUMSTD ioctl.
 */
u32 VMW_AVS_EnumVideoStandards(ENUM_STANDARD *penumstd )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED)) 
	{
		DBG("VMW_AVS_EnumVideoStandards\n");
		return DO_IOCTL(pvideo->fd_noio, VIDIOC_ENUMSTD, penumstd );
	}
	else 
	{
		MSG("VME_AVS_EnumVideoStandards: video device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_G_STD ioctl & return the format by out 
 *		parameter.
 */
u32 VMW_AVS_GetVideoStandard(VIDEO_STANDARD *pstandard )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) )
	{
		DBG("VMW_AVS_GetVideoStandard\n");
		return DO_IOCTL(pvideo->fd_noio, VIDIOC_G_STD, pstandard );
	}
	else 
	{
		MSG("VME_AVS_GetVideoStandard: video device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
	
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_S_STD ioctl.
 */
u32 VMW_AVS_SetVideoStandard(VIDEO_STANDARD *pstandard )
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) )
	{
		DBG("VMW_AVS_SetVideoStandard\n");
		return DO_IOCTL(pvideo->fd_noio, VIDIOC_S_STD, pstandard );
	}
	else 
	{
		MSG("VME_AVS_SetVideoStandard: video device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
	
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_G_TUNER ioctl & return the tuner structure
 *		with the out parameter.
 */
u32 VMW_AVT_GetTuner(VIDEO_TUNER *ptuner)
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) ) 
	{
		DBG("VMW_AVT_GetTuner\n");
		return DO_IOCTL(pvideo->fd_noio, VIDIOC_G_TUNER, ptuner );
	}
	else 
	{
		MSG("VME_AVT_GetTuner: video device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
	
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_S_TUNER ioctl.
 */
u32 VMW_AVT_SetTuner(VIDEO_TUNER *ptuner)
{
	if((pvideo != NULL) && (pvideo->status & VIDEO_OPENED) )
	{
		DBG("VMW_AVT_GetTuner\n");
		return DO_IOCTL(pvideo->fd_noio, VIDIOC_S_TUNER, ptuner );
	}
	else 
	{
		MSG("VME_AVT_SetTuner: video device not opened\n");
		return VMW_AVSE_INVSTAT;
	}
	
}

/*
 *	Algorithm:
 *		close all the open devices
 */
void VMW_AVS_VideoClose()
{
	if(pvideo == NULL){
		DBG("no device opened\n");
		return;
	}
	if(pvideo->status & VIDEO_OPENED)
		close(pvideo->fd);
		
	if(pvideo->status & VIDEO_OPENED_NOIO)
		close(pvideo->fd_noio);

	if(pvideo->status & VBI_OPENED)
		close(pvideo->fd_vbi);

	if(pvideo->status & VBI_OPENED_NOIO)
		close(pvideo->fd_vbi_noio);

	if(pvideo->status & VOUT_OPENED)
		close(pvideo->fd_vout);

	if(pvideo->status & VOUT_OPENED_NOIO)
		close(pvideo->fd_vout_noio);

	if (pvideo) free (pvideo);
	pvideo =  NULL;	
	MSG("VMW_AVS_VideoClose()");
}

/* Set the Overlay Mode. The 12x0 hardware provides 5 different
   overlay modes. These are
  
   VMW_OVL_COLOR_KEY  
     Video overlay will appear wherever graphics pixels match the
   color key comparison.
   VMW_OVL_CHROMA_KEY :
     Video overlay will appear wherever video pixels do NOT match
     the color key comparison.
   VMW_OVL_COLOR_KEY_INSIDE_ALPHA_AND_VIDEO_OUTSIDE :
     Inside alpha windows, video will appear if graphics pixels
     match the color key comparison; outside the (enabled) alpha
     windows, video will always appear.
   VMW_OVL_CHROMA_KEY_INSIDE_ALPHA_AND_GRAPHICS_OUTSIDE :
     Inside alpha windows, video will appear if video pixels do NOT match
     the color key comparison; outside the (enabled) alpha windows, video
     will not appear.
   VMW_OVL_ALWAYS_ON:
     Disables color key comparison, meaning only video is displayed
     in the video window area.
*/
u32 VMW_OVL_OverlayMode(VMW_OVL_ModeType  mode)
{
        int ret, effect = 0;
        struct v4l2_control control;
	if((pvideo != NULL) && (pvideo->status & VOUT_OPENED) ) {
                ret = ioctl(pvideo->fd_vout, VIDIOC_S_EFFECT,&effect);
                if(ret != 0) {
                        MSG("VIDIOC_S_EFFECT failed\n"); 
                        return VMW_AVSE_IOCTL;
                }
                control.id = GEODE_CID_OVERLAY_MODE;
                control.id = mode; 
                ret = ioctl(pvideo->fd_vout, VIDIOC_S_CTRL,&control);
                if(ret !=0) {
                        MSG("VIDIOC_S_CTRL failed\n"); 
                        return VMW_AVSE_IOCTL;

                }
                        		
	}
	else
	{
		MSG("Vout device not opened\n");
		return  VMW_AVSE_INVSTAT;
	}
	
}
/*
 *	Algorithm:
 *		this function sets the  color key and color mask 
 *              values.
 * 	Side Effects:
 *		also enables capture, sets graphics on video & enables 
 *		the preview 
 */
u32 VMW_OVL_ColorKey( u32 colorkey, u32 colormask)
{
	geode_general_param_t genInfo;	
        int ret;      
	if((pvideo != NULL) && (pvideo->status & VOUT_OPENED) )
	{
		genInfo.p_1 = colorkey;
		genInfo.p_2 = colormask;
		ret = ioctl( pvideo->fd_vout,GEODE_IOCTL_S_COLOR_KEY,
				&genInfo );
                if(ret != 0 ) 
                        return VMW_AVSE_IOCTL;
	}
	else
	{
		MSG("Vout device not opened\n");
		return  VMW_AVSE_INVSTAT;
	}
		
}
/*
 *	Algorithm:
 *		this function sets the enables overlay by calling 
 * 	Side Effects:
 *		also enables capture, sets graphics on video & enables 
 *		the preview 
 */
u32 VMW_OVL_EnableOverlay( )
{
	OVERLAY_INFO ovrInfo;
	geode_general_param_t genInfo;	
	
	if((pvideo != NULL) && (pvideo->status & VOUT_OPENED) )
	{
		int enable=1, ret;
		ret = ioctl(pvideo->fd_vout, VIDIOC_PREVIEW,&enable);
		if( ret != 0)
			return VMW_AVSE_IOCTL;
	}
	else
	{
		MSG("Vout device not opened\n");
		return  VMW_AVSE_INVSTAT;
	}
		
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the GEODE_IOCTL_SET_OVERLAY_INFO ioctl to disable overlay.
 */
u32 VMW_OVL_DisableOverlay()
{
	OVERLAY_INFO ovrInfo;
	int enable = 0;

	if((pvideo != NULL) && (pvideo->status & VOUT_OPENED) )
	{
		return DO_IOCTL( pvideo->fd_vout,VIDIOC_PREVIEW,
				&enable ) ;
	}
	else
	{
		MSG("Vout device not opened\n");
		return  VMW_AVSE_INVSTAT;
	}
		

}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the GEODE_IOCTL_SET_ALPHA_BLEND_INFO ioctl.
 */
u32 VMW_OVL_EnableAlphaBlendWindow( u32 startX, u32 startY,
				    u32 width , u32 height,
				    u32 alphaValue, char alphaIncValue,
				    u32 windowNumber)
{
	ALPHA_BLEND_INFO alpInfo;

	if( (pvideo != NULL) && (pvideo->status & VOUT_OPENED) )
	{
		alpInfo.TypeOfData = 2;
		alpInfo.WindowNumber = windowNumber;
		alpInfo.Priority = 1; 
		alpInfo.StartX = startX;
		alpInfo.StartY = startY;
		alpInfo.Width = width;
		alpInfo.Height = height;
		alpInfo.Alpha = alphaValue;
		alpInfo.Delta = alphaIncValue;
		alpInfo.Color = 0; 
		alpInfo.Enable = 1;


		return ( DO_IOCTL( pvideo->fd_vout,
		                GEODE_IOCTL_SET_ALPHA_BLEND_INFO,
				&alpInfo ));
	}
	else
	{
		MSG("Vout not opened\n");
		return VMW_AVSE_INVSTAT;
	}

}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the GEODE_IOCTL_SET_ALPHA_BLEND_INFO ioctl.
 */
u32 VMW_OVL_DisableAlphaWindow(u32 windowNumber) 
{
	ALPHA_BLEND_INFO alpInfo;

	if((pvideo != NULL) && ( pvideo->status & VOUT_OPENED) )
	{
		alpInfo.WindowNumber = windowNumber;
		alpInfo.Enable = 0;
		return ( DO_IOCTL( pvideo->fd_vout,
		                GEODE_IOCTL_SET_ALPHA_BLEND_INFO,
				&alpInfo ));
	}
	else
	{
		MSG("Vout not opened\n");
		return VMW_AVSE_INVSTAT;
	}
}

/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. We then
 *		call the VIDIOC_S_CTRL ioctl with proper control value passed
 *		as input parameter.
 */
u32 VMW_AVS_SetControl(u32 control, int value)
{
	VIDEO_CONTROL vcontrol;
	MSG("VMW_AVS_SetControl\n");
	if((pvideo == NULL) || !( pvideo->status & VIDEO_OPENED ) )
	{
		MSG("Video device not opened\n");
		return VMW_AVSE_INVSTAT ;
	}
	vcontrol.value = value;

	switch(control)
	{
		case AVS_CONTROL_BRIGHT:
			vcontrol.id = AVS_CONTROL_BRIGHT;
		break;

		case AVS_CONTROL_HUE:
			vcontrol.id = AVS_CONTROL_HUE;
		break;

		case AVS_CONTROL_SAT:
			vcontrol.id = AVS_CONTROL_SAT;
		break;
		/* add more as we go on */
		default:
			return VMW_AVSE_INVSTAT;
			break;
	}
	return ( DO_IOCTL( pvideo->fd, VIDIOC_S_CTRL, &vcontrol) );	
}

int VMW_AVS_GetControl(u32 control)
{
	VIDEO_CONTROL vcontrol;
	int value;
	MSG("VMW_AVS_GetControl\n");
	if((pvideo == NULL) || !( pvideo->status & VIDEO_OPENED ) )
	{
		MSG("Video device not opened\n");
		return VMW_AVSE_INVSTAT ;
	}

	switch(control)
	{
		case AVS_CONTROL_BRIGHT:
			vcontrol.id = AVS_CONTROL_BRIGHT;
		break;

		case AVS_CONTROL_HUE:
			vcontrol.id = AVS_CONTROL_HUE;
		break;

		case AVS_CONTROL_SAT:
			vcontrol.id = AVS_CONTROL_SAT;
		break;
		/* add more as we go on */
		default:
			return VMW_AVSE_INVSTAT;
			break;
	}
	DO_IOCTL( pvideo->fd, VIDIOC_G_CTRL, &vcontrol);	
	value = vcontrol.value;
	printf("MIDDLEWARE: GetControl: %d\n", vcontrol.value);
	return vcontrol.value;
}



/* added on December 18, 2000 */
/*
 *	Algorithm:
 *		if the callback function is passed as input parameter then
 *		first we register the signal handler then signal the driver
 *		to start emitting signals by GEODE_IOCTL_S_SIGNAL ioctl.
 *		otherwise we disable the signal.
 *		
 */
u32 VMW_AVV_SetupVBICallback(pVBICallback pCallBack, /*the user callback function*/
			     unsigned char * buffer,
			     int length, /* number of bytes after which to send 
			     		  * the signal */
			     void *userData /* user data buffer pointer*/
			      )
{
	signal_info_t sig_info;
	if(pCallBack != NULL)
	{
		/* set the buffers */
		pvideo->buffer = buffer;
		pvideo->buffer_length = length;
		pvideo->callback = pCallBack;
		pvideo->callback_data = userData;
	
		/* first we register the signal handler. */
		if(signal(VBI_SIGNAL_NUM,signal_handler)){
			MSG("Couldn't register signal handler\n");
			return -1;
		}
		/* now we tell the driver to start emitting signals */
		sig_info.signal_number = VBI_SIGNAL_NUM;
		sig_info.signal_on_length = length;

		if(pvideo->status & VBI_OPENED)
		{
			return DO_IOCTL(pvideo->fd_vbi,GEODE_IOCTL_S_SIGNAL,&sig_info);
		}
	}
	else /* we have to disable it */
	{
		sig_info.signal_number = 0;
		if(pvideo->status & VBI_OPENED)
		{
			return DO_IOCTL(pvideo->fd_vbi,GEODE_IOCTL_S_SIGNAL,
								&sig_info);
		}
	}
	return VMW_AVSE_INVSTAT;
}

/*--------------------------------------------------------------------------
 * Description: signal handler for the vbi data 
 * return : void
 *------------------------------------------------------------------------*/
 void signal_handler(int signum)
 {
	 int total_bytes_read = 0, bytes_read;
	 int bytes_remaining = MIN(pvideo->buffer_length,pvideo->vbi_read_buffer_length);
	 unsigned char *ptr = pvideo->buffer;

//	 MSG("In signal handler\n");
	/* the driver will signal only if it has [buffer_length] bytes
	 * ready to be read. So it is ok to read continuously till we get
	 * all the bytes.
	 */
	bytes_read = read(pvideo->fd_vbi, ptr, bytes_remaining);

	/* now we call the user specified callback function */
	pvideo->callback(bytes_read,pvideo->callback_data);
	/* it is upto the user to take care of parsing or copying this
	 * data. The next time the callback is called, the buffer will be
	 * over-written.
	 */
}


/*
 *	Algorithm:
 *		First we check if the vbi device has been opened. Then fill
 *		the buffer with information obtained as input parameters &
 *		call the GEODE_IOCTL_SET_VIDEO_BUFFER_INFO ioctl.
 */
u32 VMW_OVL_SetVIPBufferInfo(unsigned long even, unsigned long odd,
        		     unsigned long pitch)
{
	return VMW_AVSE_INVSTAT ; 
/*	
	geode_buffer_info_t buff_info;
	if(( pvideo != NULL) && pvideo->status & VIDEO_OPENED)
	{
		buff_info.odd_address = odd;
		buff_info.even_address = even;
		buff_info.pitch = pitch;

		return ioctl(pvideo->fd,GEODE_IOCTL_SET_VIDEO_BUFFER_INFO,
				&buff_info);
	}
*/	

}

/*
 *	Start or stop VBI capture. To enable VBI, this function is called
 *	with a non-zero parameter after the other VBI parameters have been
 *	established by calling VMW_AVV_SetVBIFormat, and if necessary,
 *	SetupVBICallback. We use the VIDIOC_STREAMON and VIDIOC_STREAM_OFF
 *	ioctls of the vbi device.
 */
u32 VMW_AVV_EnableVBICapture(int enable)
{
	DBG("VMW_AVV_EnableVBICapture(int)\n");
	if((pvideo != NULL) && (pvideo->status & VBI_OPENED) )
	{
		if(enable)
		{
			return
			DO_IOCTL(pvideo->fd_vbi,VIDIOC_STREAMON,&enable);	
		}
		else
		{
			return
			DO_IOCTL(pvideo->fd_vbi,VIDIOC_STREAMOFF,&enable);	
		}
	}
	else
		return VMW_AVSE_INVSTAT;
}
